# -*- coding: utf-8 -*-
"""
Common sub-module of csu_radartools; contains commonly used functions.
"""

import numpy as np

##############
#  Arrays  ###
##############


def _check_for_array(dz, zdr, kdp):
    len_flag = hasattr(dz, '__len__')
    if not len_flag:
        dz = np.array([dz])
        kdp = np.array([kdp])
        zdr = np.array([zdr])
    return dz, zdr, kdp, len_flag


################
#  Rainfall  ###
################


def calc_rain_zr(dz, a=300.0, b=1.4):
    """
    dz = Reflectivity (dBZ), returns rainfall rate (mm h^-1)
    Form Z = a * R**b
    """
    return (linearize(dz) / a)**(1.0 / b)


def calc_rain_nexrad(dz, thresh_nexrad=53.0, a=300.0, b=1.4):
    """
    dz = Reflectivity (dBZ)
    thresh_nexrad = Reflectivity cap for rain calc
    a, b => Z = a * R**b
    """
    rr = calc_rain_zr(dz, a=a, b=b)
    cond = dz > thresh_nexrad
    rr[cond] = calc_rain_zr(thresh_nexrad, a=a, b=b)
    return rr


def calc_rain_kdp_zdr(kdp, zdr, a=90.8, b=0.93, c=-0.169):
    """
    kdp = Specific Differential Phase (deg km^-1)
    zdr = Differential Reflectivity (dB)
    a, b, c = Adjustable function parameters
    """
    return a * kdp**b * 10.0**(c * zdr)


def calc_rain_z_zdr(dz, zdr, a=6.7e-3, b=0.927, c=-0.343):
    """
    dz = Reflectivity (dBZ)
    zdr = Differential Reflectivity (dB)
    a, b, c = Adjustable function parameters
    """
    return a * linearize(dz)**b * 10.0**(c * zdr)


def calc_rain_kdp(kdp, a=40.5, b=0.85):
    """
    kdp = Specific Differential Phase (deg km^-1)
    a, b = Adjustable coefficient, exponent
    """
    return a * kdp**b

########################
#  Unit Conversions  ###
########################


def dbz2z(dbz):
    """
    Convert from log [dBZ] to linear Z [mm^6 m^−3] units.

    Parameters
    ----------
    dbz : float or array
        logarithmic reflectivity value
    """
    return 10.**(np.asarray(dbz)/10.)


def linearize(dbz):
    return dbz2z(dbz)


def z2dbz(zlin):
    """
    Convert from linear Z [mm^6 m^−3] to log [dBZ] units.

    Parameters
    ----------
    zlin : float or array
        linear reflectivity units
    """
    return 10. * np.log10(np.asarray(zlin))


def si2kmh(si):
    """
    Convert from SI [m/s] wind units to km/h.

    Parameters
    ----------
    si : float or array
        Wind in SI units (m/s)
    """
    return np.asarray(si) * 3600. / 1000.


def si2mph(si):
    """
    Convert from SI wind units to miles/h [mph].

    Parameters
    ----------
    si: float or array
        Wind in SI units (m/s)
    """
    return np.asarray(si) * 0.62137 / 1000. * 3600.


def si2kts(si):
    """
    Convert from SI wind units to knots [kt].

    Parameters
    ----------
    si: float or array
        Wind in SI units (m/s)
    """
    return np.asarray(si) * 0.51


def kmh2si(kmh):
    """
    Convert from km/h to SI wind units [m/s].

    Parameters
    ----------
    kmh: float or array
        Wind in km/hr
    """
    return np.asarray(kmh) * 1000. / 3600.


def mph2si(mph):
    """
    Convert from miles/h to SI wind units [m/s].

    Parameters
    ----------
    mph: float or array
        Wind in miles per hour
    """
    return np.asarray(mph) * 1000. / (0.62137 * 3600.)


def kts2si(kts):
    """
    Convert from knots to SI wind units [m/s].

    Parameters
    ----------
    kts: float or array
        Wind in knots
    """
    return np.asarray(kts) / 0.51



# -*- coding: utf-8 -*-
"""
Functions for calculating/estimating drop size distributions from radar data.

Contacts
--------
Brenda Dolan (dolan@atmos.colostate.edu)
Timothy Lang (tjlangco@gmail.com)

References
----------
Bringi et al. (2004; JTECH) - Alternate S-band retrieval
Bringi et al. (2009; JTECH) - C-band retrieval
Bringi et al. (2013; AMS Radar Conf) - S-band retrieval
"""

from __future__ import absolute_import
from __future__ import division
import numpy as np
from warnings import warn
from .csu_liquid_ice_mass import linearize
from .common import _check_for_array

DEFAULT_MU = 3.0


def calc_dsd(dz=None, zdr=None, kdp=None, band='S', method='2013'):
    """This is the primary function to use, it calls everything else"""
    # Initialize, check for all vars, check for scalars
    if dz is None or kdp is None or zdr is None:
        warnings.warn('No dz, zdr, or kdp provided, failing ...')
        return
    dz, zdr, kdp, len_flag = _check_for_array(dz, zdr, kdp)
    # Get DSD based on wavelength
    if band == 'S':
        if method == '2013':
            d0, Nw, mu = calc_dsd_sband_bringi_2013(dz, zdr)
        else:
            d0, Nw, mu = calc_dsd_sband_bringi_2004(dz, zdr, kdp)
    elif band == 'C':
        d0, Nw, mu = calc_dsd_cband_bringi_2009(dz, zdr)
    else:
        warn('Unknown or unsupported wavelength band, try C or S')
        return
    if not len_flag:
        return d0[0], Nw[0], mu[0]
    else:
        return d0, Nw, mu


def power_law(var, a=1.0, b=1.0):
    """Generic power law equation"""
    return a * var**b

######################################################
# Bringi et al. 2009 retrievals
# Could make the following a class, but prefer to keep
# easy independent access to sub-functions
######################################################


def calc_dsd_cband_bringi_2009(dz, zdr):
    """
    Retrieves d0, Nw, and mu following the methodology of
    Bringi et al. (2009)
    Works for C-band
    """
    d0 = 0.0 * dz
    Nw = 0.0 * dz
    dz_lin = linearize(dz)
    meth1 = np.logical_and(zdr >= -0.5, zdr <= 1.25)
    d0[meth1] = d0_low_zdr(zdr[meth1])
    meth2 = zdr > 1.25
    d0[meth2] = d0_high_zdr(zdr[meth2])
    Nw = dz_lin / power_law(d0, 0.056, 7.319)
    mu = 0.0 * dz + DEFAULT_MU
    return d0, Nw, mu


def d0_low_zdr(zdr, a=0.0203, b=-0.1488, c=0.2209, d=0.5571, e=0.801):
    return power_law(zdr, a, 4) + power_law(zdr, b, 3) + \
        power_law(zdr, c, 2) + power_law(zdr, d) + e


def d0_high_zdr(zdr, a=0.0355, b=-0.3021, c=1.0556, d=0.6844):
    return power_law(zdr, a, 3) + power_law(zdr, b, 2) + power_law(zdr, c) + d

######################################################
# Bringi et al. 2013 retrievals
# Leverages Bringi et al. (2009) function design
######################################################


def calc_dsd_sband_bringi_2013(dz, zdr):
    """
    Retrieves d0, Nw, and mu following the methodology of
    Bringi et al. (2013)
    Works for S-band
    """
    d0 = 0.0 * dz
    Nw = 0.0 * dz
    dz_lin = linearize(dz)
    meth1 = zdr < 1
    d0[meth1] = d0_low_zdr(zdr[meth1], a=0.0424, b=-0.4571, c=0.6125, d=0.457,
                           e=0.8808)
    meth2 = zdr >= 1
    d0[meth2] = d0_high_zdr(zdr[meth2], a=0.0536, b=-0.1971,
                            c=0.6261, d=1.0815)
    Nw = 19.76 * dz_lin / power_law(d0, b=7.46)
    mu = 0.0 * dz + DEFAULT_MU
    return d0, Nw, mu

######################################################
# Bringi et al. 2004 retrievals
# Could make the following a class, but prefer to keep
# easy independent access to sub-functions
######################################################


def calc_dsd_sband_bringi_2004(dz, zdr, kdp):
    """
    Retrieves d0, Nw, and mu following the methodology of
    Bringi et al. (2004)
    Works for S-band
    """
    d0 = 0.0 * dz
    Nw = 0.0 * dz
    mu = 0.0 * dz
    zeta = linearize(zdr)
    dz_lin = linearize(dz)
    beta = calc_beta(dz_lin, kdp, zeta)
    # Break down DSD calculations by methods
    meth1a = np.logical_and(dz >= 35, zdr >= 0.2)
    meth1 = np.logical_and(meth1a, kdp >= 0.3)
    d0[meth1], Nw[meth1], mu[meth1] = dsd_sband_method1(
        dz_lin[meth1], kdp[meth1], zeta[meth1], beta[meth1])
    meth2 = np.logical_and(kdp < 0.3, zdr >= 0.5)
    d0[meth2], Nw[meth2], mu[meth2] = dsd_sband_method2(
        dz_lin[meth2], kdp[meth2], zdr[meth2])
    meth3a = np.logical_and(zdr >= 0.2, zdr < 0.5)
    meth3 = np.logical_and(meth3a, kdp < 0.3)
    d0[meth3], Nw[meth3], mu[meth3] = dsd_sband_method3(
        dz_lin[meth3], kdp[meth3], zdr[meth3], zeta[meth3])
    meth4a = np.logical_and(zdr >= -0.5, zdr < 0.2)
    meth4 = np.logical_and(meth4a, kdp < 0.3)
    d0[meth4], Nw[meth4], mu[meth4] = dsd_sband_method4(dz_lin[meth4],
                                                        zeta[meth4])
    return d0, Nw, mu


def calc_beta(dz, kdp, zeta, a=2.08, b=-0.365, c=0.38, d=0.965):
    """Bringi et al. (2004) Eq. A.1"""
    return a * (dz**b) * (kdp**c) * (zeta**d)


def calc_d0_sband_method1(dz, zeta, beta, a=0.56, b=0.064, c=0.024, d=-1.42):
    """Bringi et al. (2004) Eqs. A.2-A.5 """
    c1 = power_law(beta, c, d)
    return d0_from_dz_zeta(dz, zeta, a=a, b=b, c=c1)


def calc_logNw_sband_method1(dz, zeta, beta, a=3.29, b=0.058, c=-0.023,
                             d=-1.389):
    """Bringi et al. (2004) Eqs. A.6-A.9 """
    return a * (dz**b) * (zeta**(power_law(beta, c, d)))


def calc_mu_sband_method1(d0, zeta, beta, a1=203.0, a2=1.89,
                          b1=2.23, b2=0.0388,
                          c1=3.16, c2=-0.0463, d1=0.374, d2=-0.355):
    """Bringi et al. (2004) Eqs. A.10-A.14 """
    a = power_law(beta, a1, a2)
    b = power_law(beta, b1, b2)
    c = power_law(beta, c1, c2)
    d = power_law(beta, d1, d2)
    return a * ((d0**b) / (zeta-1.0)) - power_law(zeta, c, d)


def calc_d0_sband_method3(dz, zdr, zeta, a=1.81, b=0.486, c=0.6096,
                          d=0.0516, e=3.111):
    """Bringi et al. (2004) Eq. A.18"""
    part1 = ((zdr-0.2)/3.0) * power_law(zdr, a, b)
    part2 = ((0.5-zdr)/3.0) * d0_from_dz_zeta(dz, zeta)
    return part1 + part2


def calc_Nw_mult_methods(dz, d0):
    """Bringi et al. (2004) Eqs. A.17, A.19, A.21 """
    return 21.0 * dz / power_law(d0, b=7.3529)


def d0_from_dz_zeta(dz, zeta, a=0.6096, b=0.0516, c=3.111):
    """Bringi et al. (2004) Eq. A.20"""
    return a * dz**b * zeta**c


def dsd_sband_method1(dz, kdp, zeta, beta):
    """Bringi et al. (2004) Eqs. A.1-A.14 """
    d0 = calc_d0_sband_method1(dz, zeta, beta)
    logNw = calc_logNw_sband_method1(dz, zeta, beta)
    mu = calc_mu_sband_method1(d0, zeta, beta)
    return d0, 10.0**logNw, mu


def dsd_sband_method2(dz, kdp, zdr):
    """Bringi et al. (2004) Eqs. A.15-A.17 """
    d0 = power_law(zdr, 1.81, 0.486)
    Nw = calc_Nw_mult_methods(dz, d0)
    mu = 0.0 * dz + DEFAULT_MU
    return d0, Nw, mu


def dsd_sband_method3(dz, kdp, zdr, zeta):
    """Bringi et al. (2004) Eqs. A.18-A.19 """
    d0 = calc_d0_sband_method3(dz, zdr, zeta)
    Nw = calc_Nw_mult_methods(dz, d0)
    mu = 0.0 * dz + DEFAULT_MU
    return d0, Nw, mu


def dsd_sband_method4(dz, zeta):
    """Bringi et al. (2004) Eqs. A.20-A.21 """
    d0 = d0_from_dz_zeta(dz, zeta)
    Nw = calc_Nw_mult_methods(dz, d0)
    mu = 0.0 * dz + DEFAULT_MU
    return d0, Nw, mu

######################################################



"""
csu_fhc_ml1.py

Brody Fuchs, CSU, Sept 2014

Brenda Dolan, CSU October 2020
bdolan@colostate.edu

This code is part of the cold-season HCA. It is essentially scoring a fuzzy logic
algorithm just 'wet snow' or 'other'.

"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
import numpy as np
from .beta_functions import get_mbf_sets_winterML, CSV_DIR
from .calc_kdp_ray_fir import hid_beta_f

DEFAULT_WEIGHTS = {'DZ': [4, 4], 'DR': [7, 7], 'KD': [0, 0], 'RH': [14, 14],
                   'LD': [0, 0], 'T': [0, 0]}


def hid_beta(x_arr, a, b, m):
    """Beta function calculator"""
    return 1.0/(1.0 + (((x_arr - m)/a)**2)**b)


def csu_fhc_cold_newml1(use_temp=True, weights=DEFAULT_WEIGHTS,
                        method='linear', dz=None, zdr=None, ldr=None, kdp=None,
                        rho=None, T=None, verbose=False, plot_flag=False,
                        n_types=2, temp_factor=1, band='S', fdir=CSV_DIR):
    """
    This is a melting level detection algorithm.

    Arguments:
    use_temp = Set to False to not use T in HID
    weights = Dict that contains relative weights for every variable; see
              DEFAULT_WEIGHTS for expected stucture
    method = Currently support 'hybrid' or 'linear' methods; hybrid preferred
    verbose = Set to True to get text updates
    plot_flag = Flag to turn on optional beta function plots
    band = 'X', 'C', or 'S'
    temp_factor = Factor to modify depth of T effects; > 1 will broaden the
                  slopes of T MBFs
    n_types = Number of hydrometeor species
    verbose = Set to True to get text updates

    Input measurands (if not None, all must match in shape/size):
    dz = Input reflectivity scalar/array
    zdr = Input reflectivity scalar/array
    ldr = Input reflectivity scalar/array
    kdp = Input reflectivity scalar/array
    rho = Input reflectivity scalar/array
    T = Input temperature scalar/array

    Returns:
    mu = Input array + addtl dimension containing weights for each HID species
    To get dominant species number: fh = np.argmax(mu, axis=0) + 1

    HID types:           Species #:
    -------------------------------
    Other                  1
    Wet snow                     2

    """

    if dz is None:
        print('FHC fail, no reflectivity field')
        return None
    if T is None:
        use_temp = False
    # Populate fhc_vars and radar_data based on what was passed to function
    radar_data, fhc_vars, shp, sz = \
        _populate_vars(dz, zdr, kdp, rho, ldr, T, verbose)

    # Now grab the membership beta function parameters
    mbf_sets = get_mbf_sets_winterML(
        use_temp=use_temp, plot_flag=plot_flag, n_types=n_types,
        temp_factor=temp_factor, band=band, verbose=verbose, fdir=fdir)
    sets = _convert_mbf_sets(mbf_sets)

    # Check for presence of polarimetric variables
    pol_flag = _get_pol_flag(fhc_vars)

    # Check for presence of temperature
    if use_temp:
        if verbose:
            print('Using T in FHC')
    else:
        fhc_vars['T'] = 0
        if verbose:
            print('Not using T in FHC')

    # Get weighted sums
    weight_sum, varlist = _get_weight_sum(fhc_vars, weights, method, verbose)
    if weight_sum is None:
        return None

    # Now loop over every hydrometeor class
    test_list = _get_test_list(fhc_vars, weights, radar_data, sets, varlist,
                               weight_sum, pol_flag, use_temp, method, sz)
    if test_list is None:
        return None

    # Finish up
    mu = np.array(test_list)
    shp = np.concatenate([[n_types], shp])
    # print('mu',mu.shape,'sz',sz,'shp',shp)
    if verbose:
        print('mu max: ', mu.max())
    # return mu but make sure the shape is an int array
    return mu.reshape(shp.astype(np.int32))

##########################
# Private Functions Below#
##########################


def _convert_mbf_sets(mbf_sets):
    """Gets mbf_sets dict into form that matches labels used in csu_fhc"""
    sets = {}
    sets['DZ'] = mbf_sets['Zh_set']
    sets['DR'] = mbf_sets['Zdr_set']
    sets['KD'] = mbf_sets['Kdp_set']
    sets['LD'] = mbf_sets['LDR_set']
    sets['RH'] = mbf_sets['rho_set']
    sets['T'] = mbf_sets['T_set']
    return sets


def _get_pol_flag(fhc_vars):
    """Check for presence of polarimetric variables"""
    if fhc_vars['DR'] or fhc_vars['KD'] or fhc_vars['LD'] or fhc_vars['RH']:
        pol_flag = True
    else:
        pol_flag = False
    return pol_flag


def _populate_vars(dz, zdr, kdp, rho, ldr, T, verbose):

    """
    Check for presence of each var, and update dicts as needed.
    Flattens multi-dimensional arrays to optimize processing.
    The output array from csu_fhc_summer() will be re-dimensionalized later.
    """
    varlist = [dz, zdr, kdp, rho, ldr, T]
    keylist = ['DZ', 'DR', 'KD', 'RH', 'LD', 'T']
    fhc_vars = {}
    radar_data = {}
    for i, key in enumerate(keylist):
        var = varlist[i]
        if var is not None:
            if key == 'DZ':
                shp = np.shape(var)
                sz = np.size(var)
            if np.ndim(var) > 1:
                radar_data[key] = np.array(var).ravel().astype('float32')
            elif np.ndim(var) == 1:
                radar_data[key] = np.array(var).astype('float32')
            else:
                radar_data[key] = np.array([var]).astype('float32')
            fhc_vars[key] = 1
        else:
            fhc_vars[key] = 0
    if verbose:
        print('USING VARIABLES: ', fhc_vars)
    return radar_data, fhc_vars, shp, sz


def _get_weight_sum(fhc_vars, weights, method, verbose):
    """Gets sum of weights and varlist used, which depend on method"""
    if 'hybrid' in method:
        if verbose:
            print('Using hybrid HID method. Pol vars weighted,',
                  'Z and T (if used) are multiplied')
        varlist = ['DR', 'KD', 'RH', 'LD']
    elif 'linear' in method:
        if verbose:
            print('NOT using hybrid, all variables treated as weighted sum')
        varlist = ['DR', 'KD', 'RH', 'LD', 'T', 'DZ']
    else:
        print('No weighting method defined, use hybrid or linear')
        return None, None
    weight_sum = np.zeros([2])
    for i in range(0, 2):
        weight_sum[i] = np.sum(np.array([fhc_vars[key] * weights[key][i]
                                         for key in varlist]))
    if verbose:
        print('weight_sum: ', weight_sum)
    return weight_sum, varlist


def _calculate_test(fhc_vars, weights, radar_data, sets,
                    varlist, weight_sum, c, sz):
    """Loop over every var to get initial value for each HID species 'test'"""
#    test = (np.sum(np.array([fhc_vars[key] * weights[key] *
#                            hid_beta(radar_data[key], sets[key]['a'][c],
#                            sets[key]['b'][c], sets[key]['m'][c])
    test = (np.sum(np.array([fhc_vars[key] * weights[key][c] *
                            hid_beta_f(sz, radar_data[key], sets[key]['a'][c],
                            sets[key]['b'][c], sets[key]['m'][c])
            for key in varlist if key in radar_data.keys()]),
            axis=0))/weight_sum[c]
    return test


def _get_test_list(fhc_vars, weights, radar_data, sets, varlist, weight_sum,
                   pol_flag, use_temp, method, sz):
    """
    Master loop to compute HID values for each species ('test' & 'test_list').
    Depending on method used, approach is modfied.
    Currently disabling testing as it gets spoofed by bad data. Letting the
    calculations continue then mask out the bad data using other methods.
    TO DO: Change poor naming scheme for variables 'test' and 'test_list']
    """
    # TJL - Check order of if statements
    test_list = []
    for c in range(len(sets['DZ']['m'])):
        if 'hybrid' in method:  # Hybrid emphasizes Z and T extra HARD
            if pol_flag:
                test = _calculate_test(fhc_vars, weights, radar_data, sets,
                                       varlist, weight_sum, c, sz)
                # if test.max() > 1:  # Max of test should never be > 1
                #     print 'Fail loc 1, test.max() =', test.max()
                #     return None
            if use_temp:
                if pol_flag:
                    # *= multiplies by new value and stores in test
                    test *= hid_beta_f(sz, radar_data['T'], sets['T']['a'][c],
                                       sets['T']['b'][c], sets['T']['m'][c])
                    # print 'in loc 2'
                    # if test.max() > 1: #Maximum of test should never be > 1
                    #     print 'Fail loc 2, test.max() =', test.max()
                    #     return None
                else:
                    test = hid_beta_f(sz, radar_data['T'], sets['T']['a'][c],
                                      sets['T']['b'][c], sets['T']['m'][c])
            if fhc_vars['DZ']:
                if pol_flag or use_temp:
                    test *= hid_beta_f(
                        sz, radar_data['DZ'], sets['DZ']['a'][c],
                        sets['DZ']['b'][c], sets['DZ']['m'][c])
                    # if test.max() > 1:  # Max of test should never be > 1
                    #     print 'Fail loc 3, test.max() =', test.max()
                    #     return None
                else:
                    test = hid_beta_f(sz, radar_data['DZ'], sets['DZ']['a'][c],
                                      sets['DZ']['b'][c], sets['DZ']['m'][c])
        elif 'linear' in method:  # Just a giant weighted sum
            if pol_flag:
                test = _calculate_test(fhc_vars, weights, radar_data, sets,
                                       varlist, weight_sum, c, sz)
        test_list.append(test)
    return test_list



# -*- coding: utf-8 -*-
"""
Generate membership functions for hydrometeor types based on scattering models.

Construct Membership Beta Functions (MBFs) parameters for each hydrometeor
type and fuzzy set based on scattering simulations (Dolan and Rutledge, 2009).

Smaller b values means falloff is slower.

Basic form of a Membership Beta Function is::

                            1
                  ---------------------
    beta(x) =         |              |^b
                   1+ |  ((x-m)/a)^2 |
                      |              |

     Where x = input data (Zh, Zdr, etc)
           m = center
           a = width
           b = slope

Anticipate having 6 input data fields and 10 fuzzy sets (HID types)::

    Input data fields:           HID types:           Fuzzy set:
    -------------------           ----------           ----------
         Z_h                      Drizzle                  1
         Z_dr                     Rain                     2
         K_dp                     Ice Crystals             3
         LDR                      Aggregates               4
         rho_hv                   Wet Snow                 5
         Temperature              Vertical Ice             6
                                  Low Density Graupel      7
                                  High Density Graupel     8
                                  Hail                     9
                                  Big Drops               10

Contact
-------

Kyle C. Wiens
Brenda Dolan
bdolan@atmos.colostate.edu
26 April 2002.
"""

from __future__ import division
from __future__ import print_function
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os

# Beta function parameters stored in individual CSVs in separate directory
CSV_DIR = os.sep.join([os.path.dirname(__file__),
                      'beta_function_parameters'])+'/'

################################
# Helper Functions Below
################################


def _get_beta_example(var, mbf_sets, field, i):
    """Internal beta function calculator"""
    return 1.0/(1.0 + (((var - mbf_sets[field]['m'][i]) /
                mbf_sets[field]['a'][i])**2)**mbf_sets[field]['b'][i])


def plot_beta_functions(mbf_sets, band, temp_factor, pdir='./', n_types=10):
    """
    Optional plotting routine that produces basic beta function curves
    for every hydrometeor species
    mbf_sets = Membership beta function sets
    band = 'X', 'C' or 'S'
    temp_factor = Factor to modify depth of T effects
    fdir = Path to plot locations
    n_types = Number of hydrometeor species
    """
    label = ['Drizzle', 'Rain', 'Ice_Crystals', 'Aggregates',
             'Wet_Snow', 'Vertical_Ice', 'Low-Density_Graupel',
             'High-Density_Graupel', 'Hail', 'Big_Drops']
    # 'Large_Hail','SmHail_Rain','LgHail_Rain']
    # DUMMY ARRAYS TO PLOT OVER
    dz = np.arange(701)/10.
    zdr = -2.0 + np.arange(801)/100.
    kdp = -2.0 + np.arange(901)/100.
    ldr = -60.0 + np.arange(5900)/100.
    rho_hv = 0.75 + np.arange(2501)/10000.
    T = -50. + np.arange(1001)/10.
    for i in range(n_types):
        dz_beta = _get_beta_example(dz, mbf_sets, 'Zh_set', i)
        kdp_beta = _get_beta_example(kdp, mbf_sets, 'Kdp_set', i)
        zdr_beta = _get_beta_example(zdr, mbf_sets, 'Zdr_set', i)
        ldr_beta = _get_beta_example(ldr, mbf_sets, 'LDR_set', i)
        rho_beta = _get_beta_example(rho_hv, mbf_sets, 'rho_set', i)
        T_beta = _get_beta_example(T, mbf_sets, 'T_set', i)
        fig, ax = plt.subplots(3, 2, figsize=(8, 5))
        fig.subplots_adjust(top=0.90, bottom=0.1, left=0.05, right=0.95,
                            hspace=0.45)
        ax = ax.flatten()
        fig.suptitle('%s band %s, temp_factor: %.1f' % (band,
                     label[i], temp_factor))
        ax[0].plot(dz, dz_beta, linewidth=2)
        ax[0].set_title('dBZ')
        ax[1].plot(kdp, kdp_beta, linewidth=2)
        ax[1].set_title('Kdp')
        ax[2].plot(zdr, zdr_beta, linewidth=2)
        ax[2].set_title('Zdr')
        ax[3].plot(ldr, ldr_beta, linewidth=2)
        ax[3].set_title('LDR')
        ax[3].set_ylim(0, 1)
        ax[4].plot(rho_hv, rho_beta, linewidth=2)
        ax[4].set_title(r'$\rho$$_{hv}$')
        ax[5].plot(T, T_beta, linewidth=2)
        ax[5].set_title('T')
        plt.savefig(pdir+'%s_beta_%s.png' % (label[i], band))
        plt.close()


def get_beta_set(filename, factor=1.0):
    """
    Function to read the CSVs that contain the beta function parameters for
    each variable and frequency band
    filename = Name of file
    factor = Modifier to effect of variable (normally only used for T)
    """
    beta_parameters = pd.read_csv(filename, index_col=0)
    set_values = {}
    for label in ['m', 'a', 'b']:
        set_values[label] = np.array(beta_parameters.loc[:, label]) / factor
    return set_values

################################
def get_mbf_sets_winter(use_temp=False, plot_flag=False, n_types=2,
                        temp_factor=1, band='S', fdir=CSV_DIR, verbose=False,
                        pdir='./'):
    """
    Define structures to hold the beta function parameters for each input
    variable and fuzzy set. Valid for summer precip.

    HID types:           Fuzzy set:
    -------------------------------
    other                  1
    ML                      2

    Arguments:
    use_temp = Flag to create T beta function sets
    plot_flag = Flag to turn on optional beta function plots
    band = 'C' or 'S'
    temp_factor = Factor to modify depth of T effects; > 1 will broaden the
                  slopes of T MBFs
    fdir = Path to CSV locations
    n_types = Number of hydrometeor species
    verbose = Set to True to get text updates
    pdir = Directory to stash beta function plots (if made)

    Returns:
    mbf_sets = Membership beta function sets
    """

    me = '-->get_mbf_sets_winter  '
    if verbose:
        print(me + ' 2 Category ML %s-band HID' % band)

    # Horizontal Reflectivity (Zh)
    fname = band + '-band_Reflectivity_winter.csv'
    Zh_set = get_beta_set(fdir+fname)

    # Differential Reflectivity (Zdr)
    fname = band + '-band_Differential_Reflectivity_winter.csv'
    Zdr_set = get_beta_set(fdir+fname)

    # Specific Differential Phase (Kdp)
    fname = band + '-band_Specific_Differential_Phase_winter.csv'
    Kdp_set = get_beta_set(fdir+fname)

    # Linear depolarization ratio (LDR)
    fname = band + '-band_Linear_Depolarization_Ratio_winter.csv'
    LDR_set = get_beta_set(fdir+fname)

    # Correlation coefficient at zero lag (rho)
    fname = band + '-band_Correlation_Coefficient_winter.csv'
    rho_set = get_beta_set(fdir+fname)

    # Temperature (T)
    if use_temp:
        fname = band + '-band_Temperature_winter.csv'
        T_set = get_beta_set(fdir+fname, factor=temp_factor)
        if verbose:
            print(me+'Using Temperature in HID')
    else:
        T_set = None
        if verbose:
            print(me+'No Temperature used in HID')

    # Now populate the full mbf_sets dictionary
    mbf_sets = {'Zh_set': Zh_set, 'Zdr_set': Zdr_set, 'Kdp_set': Kdp_set,
                'LDR_set': LDR_set, 'rho_set': rho_set, 'T_set': T_set}

    if plot_flag:
        plot_beta_functions(mbf_sets, band, temp_factor, pdir=pdir,
                            n_types=n_types)

    return mbf_sets
    
def get_mbf_sets_winterML(use_temp=False, plot_flag=False, n_types=2,
                        temp_factor=1, band='S', fdir=CSV_DIR, verbose=False,
                        pdir='./'):
    """
    Define structures to hold the beta function parameters for each input
    variable and fuzzy set. Valid for summer precip.

    HID types:           Fuzzy set:
    -------------------------------
    other                  1
    ML                      2

    Arguments:
    use_temp = Flag to create T beta function sets
    plot_flag = Flag to turn on optional beta function plots
    band = 'C' or 'S'
    temp_factor = Factor to modify depth of T effects; > 1 will broaden the
                  slopes of T MBFs
    fdir = Path to CSV locations
    n_types = Number of hydrometeor species
    verbose = Set to True to get text updates
    pdir = Directory to stash beta function plots (if made)

    Returns:
    mbf_sets = Membership beta function sets
    """

    me = '-->get_mbf_sets_summer  '
    if verbose:
        print(me + ' 2 Category ML %s-band HID' % band)

    # Horizontal Reflectivity (Zh)
    fname = band + '-band_Reflectivity_ML.csv'
    Zh_set = get_beta_set(fdir+fname)

    # Differential Reflectivity (Zdr)
    fname = band + '-band_Differential_Reflectivity_ML.csv'
    Zdr_set = get_beta_set(fdir+fname)

    # Specific Differential Phase (Kdp)
    fname = band + '-band_Specific_Differential_Phase_ML.csv'
    Kdp_set = get_beta_set(fdir+fname)

    # Linear depolarization ratio (LDR)
    fname = band + '-band_Linear_Depolarization_Ratio_ML.csv'
    LDR_set = get_beta_set(fdir+fname)

    # Correlation coefficient at zero lag (rho)
    fname = band + '-band_Correlation_Coefficient_ML.csv'
    rho_set = get_beta_set(fdir+fname)

    # Temperature (T)
    if use_temp:
        fname = band + '-band_Temperature_ML.csv'
        T_set = get_beta_set(fdir+fname, factor=temp_factor)
        if verbose:
            print(me+'Using Temperature in HID')
    else:
        T_set = None
        if verbose:
            print(me+'No Temperature used in HID')

    # Now populate the full mbf_sets dictionary
    mbf_sets = {'Zh_set': Zh_set, 'Zdr_set': Zdr_set, 'Kdp_set': Kdp_set,
                'LDR_set': LDR_set, 'rho_set': rho_set, 'T_set': T_set}

    if plot_flag:
        plot_beta_functions(mbf_sets, band, temp_factor, pdir=pdir,
                            n_types=n_types)

    return mbf_sets

def get_mbf_sets_summerML(use_temp=False, plot_flag=False, n_types=2,
                        temp_factor=1, band='S', fdir=CSV_DIR, verbose=False,
                        pdir='./'):
    """
    Define structures to hold the beta function parameters for each input
    variable and fuzzy set. Valid for summer precip.
    def get_mbf_sets_winter(use_temp=False, plot_flag=False, n_types=2,
                        temp_factor=1, band='S', fdir=CSV_DIR, verbose=False,
                        pdir='./'):
    Define structures to hold the beta function parameters for each input
    variable and fuzzy set. Valid for summer precip.

    HID types:           Fuzzy set:
    -------------------------------
    other                  1
    ML                      2

    Arguments:
    use_temp = Flag to create T beta function sets
    plot_flag = Flag to turn on optional beta function plots
    band = 'C' or 'S'
    temp_factor = Factor to modify depth of T effects; > 1 will broaden the
                  slopes of T MBFs
    fdir = Path to CSV locations
    n_types = Number of hydrometeor species
    verbose = Set to True to get text updates
    pdir = Directory to stash beta function plots (if made)

    Returns:
    mbf_sets = Membership beta function sets
    """

    me = '-->get_mbf_sets_summerML  '
    if verbose:
        print(me + ' 2 Category ML %s-band HID' % band)

    # Horizontal Reflectivity (Zh)
    fname = band + '-band_Reflectivity_winter.csv'
    Zh_set = get_beta_set(fdir+fname)

    # Differential Reflectivity (Zdr)
    fname = band + '-band_Differential_Reflectivity_winter.csv'
    Zdr_set = get_beta_set(fdir+fname)

    # Specific Differential Phase (Kdp)
    fname = band + '-band_Specific_Differential_Phase_winter.csv'
    Kdp_set = get_beta_set(fdir+fname)

    # Linear depolarization ratio (LDR)
    fname = band + '-band_Linear_Depolarization_Ratio_winter.csv'
    LDR_set = get_beta_set(fdir+fname)

    # Correlation coefficient at zero lag (rho)
    fname = band + '-band_Correlation_Coefficient_winter.csv'
    rho_set = get_beta_set(fdir+fname)

    # Temperature (T)
    if use_temp:
        fname = band + '-band_Temperature_winter.csv'
        T_set = get_beta_set(fdir+fname, factor=temp_factor)
        if verbose:
            print(me+'Using Temperature in HID')
    else:
        T_set = None
        if verbose:
            print(me+'No Temperature used in HID')

    # Now populate the full mbf_sets dictionary
    mbf_sets = {'Zh_set': Zh_set, 'Zdr_set': Zdr_set, 'Kdp_set': Kdp_set,
                'LDR_set': LDR_set, 'rho_set': rho_set, 'T_set': T_set}

    if plot_flag:
        plot_beta_functions(mbf_sets, band, temp_factor, pdir=pdir,
                            n_types=n_types)

    return mbf_sets
    
    

def get_mbf_sets_summer(use_temp=True, plot_flag=False, n_types=10,
                        temp_factor=1, band='S', fdir=CSV_DIR, verbose=False,
                        pdir='./'):
    """
    Define structures to hold the beta function parameters for each input
    variable and fuzzy set. Valid for summer precip.

    HID types:           Fuzzy set:
    -------------------------------
    Drizzle                  1
    Rain                     2
    Ice Crystals             3
    Aggregates               4
    Wet Snow                 5
    Vertical Ice             6
    Low-Density Graupel      7
    High-Density Graupel     8
    Hail                     9
    Big Drops                10

    Arguments:
    use_temp = Flag to create T beta function sets
    plot_flag = Flag to turn on optional beta function plots
    band = 'C' or 'S'
    temp_factor = Factor to modify depth of T effects; > 1 will broaden the
                  slopes of T MBFs
    fdir = Path to CSV locations
    n_types = Number of hydrometeor species
    verbose = Set to True to get text updates
    pdir = Directory to stash beta function plots (if made)

    Returns:
    mbf_sets = Membership beta function sets
    """

    me = '-->get_mbf_sets_summer  '
    if verbose:
        print(me + ' 10 Category Summer %s-band HID' % band)

    # Horizontal Reflectivity (Zh)
    fname = band + '-band_Reflectivity.csv'
    Zh_set = get_beta_set(fdir+fname)

    # Differential Reflectivity (Zdr)
    fname = band + '-band_Differential_Reflectivity.csv'
    Zdr_set = get_beta_set(fdir+fname)

    # Specific Differential Phase (Kdp)
    fname = band + '-band_Specific_Differential_Phase.csv'
    Kdp_set = get_beta_set(fdir+fname)

    # Linear depolarization ratio (LDR)
    fname = band + '-band_Linear_Depolarization_Ratio.csv'
    LDR_set = get_beta_set(fdir+fname)

    # Correlation coefficient at zero lag (rho)
    fname = band + '-band_Correlation_Coefficient.csv'
    rho_set = get_beta_set(fdir+fname)

    # Temperature (T)
    if use_temp:
        fname = band + '-band_Temperature.csv'
        T_set = get_beta_set(fdir+fname, factor=temp_factor)
        if verbose:
            print(me+'Using Temperature in HID')
    else:
        T_set = None
        if verbose:
            print(me+'No Temperature used in HID')

    # Now populate the full mbf_sets dictionary
    mbf_sets = {'Zh_set': Zh_set, 'Zdr_set': Zdr_set, 'Kdp_set': Kdp_set,
                'LDR_set': LDR_set, 'rho_set': rho_set, 'T_set': T_set}

    if plot_flag:
        plot_beta_functions(mbf_sets, band, temp_factor, pdir=pdir,
                            n_types=n_types)

    return mbf_sets


################################

if __name__ == '__main__':
    a = cdf_betas_sum(plot_flag=True, use_temp=True, temp_factor=2, band='C')



# -*- coding: utf-8 -*-
"""
csu_misc sub-module of csu_radartools.

Contacts
--------
Brenda Dolan (dolan@atmos.colostate.edu)
Paul Hein (hein@atmos.colostate.edu)
Timothy Lang (tjlangco@gmail.com)

References
----------
Lang et al. (2007; JCLIM) - Insect filter
"""
from __future__ import division
import numpy as np
from warnings import warn
from scipy.ndimage.measurements import label

VERSION = '1.3'

# Biological scatterer thresholds originally created using NAME data
DEFAULT_DZ_RANGE = [[-100, 10], [10, 15], [15, 20],
                    [20, 25], [25, 30], [30, 35]]
DEFAULT_DR_THRESH = [1, 1.3, 1.7, 2.1, 2.5, 2.8]


def insect_filter(dz, zdr, height=None, mask=None, dz_range=DEFAULT_DZ_RANGE,
                  dr_thresh=DEFAULT_DR_THRESH, bad=-32768):
    """
    Returns a mask that identifies potentially suspect gates due to presence of
    biological scatterers.
    """
    # Define generic mask if none provided
    if mask is None:
        mask = np.logical_or(dz == bad, zdr == bad)
    cond = []
    for i, thresh in enumerate(dr_thresh):
        dz_subrange = dz_range[i]
        cond_dz = np.logical_and(dz > dz_subrange[0], dz <= dz_subrange[1])
        sub_cond = np.logical_and(cond_dz, zdr >= thresh)
        cond.append(sub_cond)
    store_cond = mask
    for sub_cond in cond:
        store_cond = np.logical_or(store_cond, sub_cond)
    return store_cond


def second_trip_filter_magnetron(vr, bad=-32768):
    """
    With magnetron transmitters, there is no coherent Doppler signature in
    second trip. Thus, if velocity is missing but there is echo, that echo is
    likely second trip.
    """
    return vr == bad


def differential_phase_filter(sdp, thresh_sdp=12):
    """
    Filter out gates with large standard deviations of differential phase.
    Note: thresh_sdp can be an array the same size as sdp.
    """
    return sdp > thresh_sdp


def despeckle(data, bad=-32768, ngates=4):
    """
    data = 1-D or 2-D array of radar data (e.g., reflectivity) to despeckle.
           If 1-D, should be a single ray. If 2-D, first dimension should be
           azimuth/elevation and second should be range.
    bad = Bad data value to check against
    ngates = Number of contiguous good data gates required along ray for no
             masking to occur

    Returns mask w/ same shape as data. True = bad.
    """
    if not hasattr(data, '__len__'):
        warn('Does not work with scalars, #Failing ...')
        return
    array = 1.0 * data
    mask = array == bad
    array[mask] = 0
    array[~mask] = 1
    if np.ndim(array) == 2:
        for az in np.arange(np.shape(array)[0]):
            mask[az, :] = _despeckle_ray(array[az, :], bad, ngates)
    elif np.ndim(array) == 1:
        mask = _despeckle_ray(array, bad, ngates)
    else:
        warn('Need 1- or 2-D array, #Failing ...')
        return
    return mask


def _despeckle_ray(ray, bad, ngates):
    """
    Arguments similar to despeckle(). Should only be sent rays.
    """
    lab, nf = label(ray)
    if nf > 0:
        labels = lab[lab > 0]
        for feat in np.linspace(np.min(labels), np.max(labels),
                                num=np.max(labels)-np.min(labels)+1):
            featsz = np.size(labels[labels == feat])
            if featsz <= ngates:
                lab[lab == feat] = 0
    return lab == 0



# -*- coding: utf-8 -*-
"""
fundamentals sub-module of csu_radartools

Functions to calculate:
    radar system characteristics (System)
    Doppler radar system characteristics (Doppler)
    radar geometry characteristics (Geometry)
    radar-derived variables (Variables)
    coefficients used with attenuation calculations (Attenuation)

References
----------
Battan (1973), Radar Observations of the Atmosphere.
Doviak and Zrnic (1993), Doppler Radar and Weather Observations.
Rinehart (1997), Radar for Meteorologists.
Jorgensen (1983; JCAM), Feasibility Test of an Airborne Pulse-Doppler
    Meteorological Radar.
Bech et al. (2003; JAOT), The Sensitivity of Single Polarization Weather Radar
    Beam Blockage Correction to Variability in the Vertical Refractivity
    Gradient.
Aydin et al. (1986; JCAM), Remote Sensing of Hail with Dual Linear
    Polarization Radar

Nick Guy - 10 May 2016, Ported from PyRadarMet
"""
import numpy as np

speed_of_light = 3e8  # Speed of light [m/s]
k_boltz = 1.381e-23  # Boltzmann's constant [ m^2 kg s^-2 K^-1]
earth_radius = 6371000.  # Earth's average radius [m] assuming sphericity
r43 = earth_radius * 4./3.
# Earth radius according to International Union of Geodesy and Geophysics
# r43 is 4/3 Approximation effective radius for standard atmosphere [m]

############
#  System  #
############


def antenna_gain(p1, p2):
    """
    Antenna gain [dB] via power ratio.

    From Rinehart (1997), Eqn 2.1

    Parameters
    ----------
    p1 : float or array
        Power on the beam axis [W]
    p2 : float or array
        Power from an isotropic antenna [W]

    Notes
    -----
    Ensure that both powers have the same units!
    If arrays are used, either for one okay, or both must be the same length.
    """
    return 10. * np.log10(np.asarray(p1) / np.asarray(p2))


def frequency(wavelength):
    """
    Frequency [Hz] given wavelength.

    Parameters
    ----------
    wavelength : float or array
        Wavelength [m]
    """
    return speed_of_light / np.asarray(wavelength)


def wavelength(freq):
    """
    Wavelength [m] given frequency.

    Parameters
    ----------
    freq : float or array
        Frequency [Hz]
    """
    return speed_of_light / np.asarray(freq)


def pulse_length(pdur):
    """
    Pulse length [m] given pulse duration.

    Parameters
    ----------
    pDur : float or array
        Pulse duration [s]

    Notes
    -----
    This equation is only interested in return pulses to radar, leading
    to the factor of 1/2.
    """
    return speed_of_light * np.asarray(pdur) / 2.


def pulse_duration(tau):
    """
    Pulse duration [s] from pulse length.

    Parameters
    ----------
    tau : float or array
        Pulse length [m]
    """
    return 2 * np.asarray(tau) / speed_of_light


def radar_const(power_t, gain, tau, wavelength, bw_h, bw_v, aloss, rloss):
    """
    Radar constant. Unitless.

    From CSU Radar Meteorology notes, AT 741

    Parameters
    ----------
    power_t : float
        Transmitted power [W]
    gain : float
        Antenna Gain [dB]
    tau : float
        Pulse Width [s]
    wavelength : float
        Radar wavelength [m]
    bw_h : float
        Horizontalntenna beamwidth [degrees]
    bw_v : float
        Vertical antenna beamwidth [degrees]
    aloss : float
        Antenna/waveguide/coupler loss [dB]
    rloss : float
        Receiver loss [dB]
    """
    # Convert from dB to linear units
    alosslin = 10**(aloss / 10.)
    rlosslin = 10**(rloss / 10.)
    gainlin = 10**(gain / 10.)

    # Convert beamwidth to radians
    bw_hr = np.deg2rad(bw_h)
    bw_vr = np.deg2rad(bw_v)

    # Calculate the numerator
    numer = (np.pi**3 * speed_of_light * power_t * gainlin**2 * tau *
             bw_hr * bw_vr * alosslin * rlosslin)

    # Calculate the denominator
    denom = 1024. * np.log(2) * wavelength**2
    return numer/denom


def ant_eff_area(gain, wavelength):
    """
    Antenna effective area. [m^-2]

    From Rinehart (1997), Eqn 4.5

    Parameters
    ----------
    gain : float or array
        Antenna Gain [dB]
    wavelength : float
        Radar wavelength [m]
    """
    # Convert from dB to linear units
    gainlin = 10**(np.asarray(gain) / 10.)
    return gainlin * wavelength**2 / (4 * np.pi)


def power_target(power_t, gain, areat, range):
    """
    Power [W] intercepted by target.

    From Rinehart (1997), Eqn 4.3

    Parameters
    ----------
    power_t : float
        Transmitted power [W]
    gain : float
        Antenna gain [dB]
    areat : float
        Area of target [m^2]
    range : float or array
        Distance to sample volume from radar [m]
    """
    # Convert from dB to linear units
    gainlin = 10**(gain / 10.)
    return (power_t * gainlin * areat) / (4 * np.pi * np.asarray(range)**2)


def xsec_bscatter_sphere(diam, wavelength, dielectric=0.93):
    """
    Backscatter cross-sectional area [m^-2] of a sphere using
    the Rayleigh approximation.

    From Rinehart (1997), Eqn 4.9 and 5.7

    Parameters
    ----------
    diam : float or array
        Diamter of target [m]
    wavelength : float
        Radar wavelength [m]
    dielectric : float
        Dielectric factor [unitless]

    Notes
    -----
    The Rayleigh approximation is good when the diamter of a spherical particle
    is much smaller than the wavelength of the radar (D/wavelength= 1/16).
    This condition leads to the relationship that the area is proportional
    to the sixth power of the diameter.

    The default is for a dielectric factor value for water.  This can be
    changed by the user, e.g. K=0.208 for particle sizes of equivalent melted
    diameters or K=0.176 for particle sizes of equivalent ice spheres.
    """
    return (np.pi**5 * dielectric**2 * np.asarray(diam)**6) / wavelength**4


def norm_xsec_bscatter_sphere(diam, wavelength, dielectric=0.93):
    """
    Normalized Backscatter cross-sectional area [m^2] of a sphere using
    the Rayleigh approximation.

    From Rinehart (1997), Eqn 4.9 and 5.7 and Battan Ch. 4.5

    Parameters
    ----------
    diam : float or array
        Diamter of targer [m]
    wavelength : float
        Radar wavelength [m]
    dielectric : float
        Dielectric factor [unitless]

    Notes
    -----
    The Rayleigh approximation is good when the diamter of a spherical particle
    is much smaller than the wavelength of the radar (D/wavelength= 1/16).
    This condition leads to the relationship that the area is proportional
    to the sixth power of the diameter.

    The default is for a dielectric factor value for water.  This can be
    changed by the user, e.g. K=0.208 for particle sizes of equivalent melted
    diameters or K=0.176 for particle sizes of equivalent ice spheres.
    """

    # Calculate the cross-sectional backscatter area
    sig = xsec_bscatter_sphere(np.asarray(diam), wavelength, dielectric)
    return sig / (np.pi * (np.asarray(diam)/2.)**2)


def size_param(diam, wavelength):
    """
    Size parameter calculation. Unitless.

    From Rinehart (1997), Eqn 4.9 and 5.7 and Battan Ch. 4.5

    Parameters
    ----------
    diam : float or float
        Diamter of target [m]
    wavelength : float
        Radar wavelength [m]

    Notes
    -----
    The size paramter can be used along with the backscattering cross-section
    to distinguish ice and water dielectric characteristics.

    For example:
    Alpha < 2 the backscattering cross-section of ice is smaller than water,
    Alpha > 2 the opposite is true due to the fact that absorption in water
    exceeds that in ice.
    """
    return 2 * np.pi * np.asarray(diam)/2. / wavelength


def power_return_target(power_t, gain, wavelength, sig, range):
    """
    Power [W] returned y target located at the center of the antenna
    beam pattern.

    From Rinehart (1997), Eqn 4.7

    Parameters
    ----------
    power_t : float
        Transmitted power [W]
    gain : float
        Antenna gain [dB]
    wavelength : float
        Radar wavelength [m]
    sig : float
        Backscattering cross-sectional area of target [m^2]
    range : float or array
        Distance to sample volume from radar [m]
    """
    # Convert from dB to linear units
    gainlin = 10**(gain/10.)
    return ((power_t * gainlin**2 * wavelength**2 * sig) /
            (64 * np.pi**3 * np.asarray(range)**4))


def thermal_noise(bandwidth, units, noise_temp=290.):
    """
    Thermal noise power [W or 'dBm'].

    From CSU Radar Meteorology notes, AT741

    Parameters
    ----------
    bandwidth : float or array
        Receiver bandwidth [Hz]
    units : float
        String of nits desired, can be 'W' or 'dBm'
    noise_temp : float
        Reciever noise temperature [K]

    Notes
    -----
    Reciever noise temp set to conventional 290K by default
    """
    # Calculate the noise, convert if requested
    noise = k_boltz * noise_temp * np.asarray(bandwidth)

    if units.upper() == 'W':
        noiset = noise
    elif units.upper() == 'DBM':
        noiset = 10. * np.log10(noise/10**-3)
    else:
        print("Units must be in 'W' or 'dBm'")
        noiset = np.nan
    return noiset

#############
#  Doppler  #
#############


def fmax(prf):
    """Maximum frequency [Hz] given PRF.

    From Rinehart (1997), Eqn 6.8

    Parameters
    ----------
    prf : float or array
        Pulse repetition frequency [Hz]
    """
    return np.asarray(prf) / 2.


def vmax(prf, wavelength):
    """Nyquist velocity, or maximum unambiguous Doppler velocity (+ or -) [m/s].

    From Rinehart (1997), Eqn 6.7

    Parameters
    ----------
    prf : float or array
        Radar pulse repetition frequency [Hz]
    wavelength : float or array
        Radar wavelength [m]
    """
    return np.asarray(prf) * wavelength / 4.


def nyquist(prf, wavelength):
    """Wrapper function for vmax."""
    return vmax(prf, wavelength)


def rmax(prf):
    """Maximum unamiguous range [m].

    From Rinehart (1997), Eqn 6.11

    Parameters
    ----------
    prf : float or array
        Pulse repetition frequency [Hz]
    """
    return speed_of_light / (2. * np.asarray(prf))


def unambiguous_range(prf):
    """Wrapper for rmax."""
    return rmax(prf)


def doppler_dilemma(varin, wavelength):
    """
    The "Doppler dilemma" is the fact that both the Nyquist velocity and
    unambiguous maximum range of the radar are based upon the PRF of
    the system.

    However, they are inversely proportional, meaning that increasing one
    requires a decrease in the other.  A trade-off inherent in Doppler radar
    systems.  This relationship allows a solution for one variable given the
    value of the other.

    From Rinehart (1997), Eqn 6.12

    Parameters
    ----------
    varin : float or array
        Nyquist Velocity [m/s] or Maximum unambiguous range [m]
    wavelength : float
        Radar wavelength [m]
    """
    return (speed_of_light * wavelength / 8.) / np.asarray(varin)

    ######################
    #  Mobile Platforms  #
    ######################


def vel_shift(ground_speed, psi):
    """
    Adjusted Doppler velocity [m/s] from a mobile platform.
    Shift in Doppler velocity from mobile perspective.

    Jorgensen (1983), Eqn 2

    Parameters
    ----------
    ground_speed : float or array
        Gound speed [m/s]
    psi : float or array
        Angle between actual azimuth and fore/aft angle [deg]

    Notes
    -----
    In the case of a mobile platform (such as the NOAA P-3 aircraft, the
      Doppler velocity must be adjusted for movement of the scanning platform.

    The fore/aft angle is defined as the angle fore or aft from a plane
      normal to the direction of motion
    """
    len_gs = len(np.asarray(ground_speed))
    len_psi = len(np.asarray(psi))
#    if  len_gs != len_psi and len_gs != 1 and len_psi != 1:
    return np.asarray(ground_speed) * np.cos(np.deg2rad(psi))


def vmax_dual(wavelength, prf1, prf2):
    """Doppler velocity [m/s] from dual PRF scheme radar (+ or -).

    From Jorgensen (1983), Eqn 2

    Parameters
    ----------
    wavelength : float
        Radar wavelength [m]
    prf1 : float
        First Pulse repetition frequency [Hz]
    prf2 : float
        Second Pulse repetition frequency [Hz]

    Notes
    -----
    In the case of a mobile platform (such as the NOAA P-3 aircraft, the
    Doppler velocity must be adjusted for movement of the scanning platform.

    The fore/aft angle is defined as the angle fore or aft from a plane
    normal to the direction of motion
    """
    return (wavelength / (4 * ((1. / prf1) - (1. / prf2))))


def dual_nyquist(wavelength, prf1, prf2):
    """Wrapper for vmax_dual."""
    return vmax_dual(wavelength, prf1, prf2)


##############
#  Geometry  #
##############


def r_effective(dndh=-39e-6):
    """
    Effective radius [m] calculation.

    Rinehart (1997), Eqn 3.9, solved for R'

    Parameters
    ----------
    dndh : float
        Refraction [N x10^-6/km]

    Notes
    -----
    Effective radius of earth given a refraction.  If no refraction is given
    a "standard atmosphere" is assumed, the valued needed to have straight
    radar rays.
    """
    # Convert earth's radius to km for common dN/dH values and then
    # multiply by 1000 to return radius in meters
    return (1. / ((1/(earth_radius/1000.)) + (dndh))) * 1000.


def half_power_radius(range, bwhalf):
    """
    Half-power radius [m].

    Battan (1973),

    Parameters
    ----------
    range : float or array
        Range [m]
    bwhalf : float
        Half-power beam width [degrees]
    """
    # Convert earth's radius to km for common dN/dH values and then
    # multiply by 1000 to return radius in meters
    return (np.asarray(range) * np.deg2rad(bwhalf)) / 2.


def ray_height(range, elev, h0, reff=r43):
    """
    Center of radar beam height [m] calculation.

    Rinehart (1997), Eqn 3.12, Bech et al. (2003) Eqn 3

    Parameters
    ----------
    range : float or array
        Range from radar to point of interest [m]
    elev : float
        Elevation angle of radar beam [deg]
    h0 : float
        Height of radar antenna [m]
    reff : float
        Effective radius

    Notes
    -----
    If no Effective radius is given a "standard atmosphere" is assumed,
    the 4/3 approximation.

    Bech et al. (2003) use a factor ke that is the ratio of earth's radius
    to the effective radius (see r_effective function) and Eqn 4 in B03
    """
    # Convert earth's radius to km for common dN/dH values and then
    # multiply by 1000 to return radius in meters
    term1 = (np.sqrt(np.asarray(range)**2 + reff**2 +
             2 * np.asarray(range) * reff * np.sin(np.deg2rad(elev))))
    h = term1 - reff + h0
    return h


def sample_vol_ideal(r, bw_h, bw_v, pulse_length):
    """
    Idealized Sample volume [m^3] assuming all power in half-power beamwidths.

    From Rinehart (1997), Eqn 5.2

    Parameters
    ----------
    r : float or array
        Distance to sample volume from radar [m]
    bw_h : float
        Horizontal beamwidth [deg]
    bw_v : float
        Vertical beamwidth deg]
    pulse_length : float
        Pulse length [m]

    Notes
    -----
    This form assumes all transmitted energy is in the half-power beamwidths.
    A more realistic solution is found in the sample_vol_gauss function
    """
    return (np.pi * (np.asarray(r) * np.deg2rad(bw_h)/2.) * (np.asarray(r) *
            np.deg2rad(bw_v)/2.) * (pulse_length/2.))


def sample_vol_gauss(range, bw_h, bw_v, pulse_length):
    """
    Sample volume [m^3] assuming transmitted energy in Gaussian beam shape.

    From Rinehart (1997), Eqn 5.4

    Parameters
    ----------
    range  : float or array
        Distance to sample volume from radar [m]
    bw_h : float
        Horizontal beamwidth [deg]
    bw_v : float
        Vertical beamwidth deg]
    pulse_length : float
        Pulse length [m]

    Notes
    -----
    This form assumes a Gaussian beam shape for transmitted energy and is more
    realistic than the sample_vol_ideal.  Derived by Probert-Jones (1962).
    """
    numer = (np.pi * np.asarray(range)**2 * np.deg2rad(bw_h) *
             np.deg2rad(bw_v) * pulse_length)
    denom = 16. * np.log(2)

    SVol = numer / denom
    return SVol


def range_correct(range, height, elev):
    """
    A corrected range [m] from radar that takes into account the "loss" of
    ground distance because of the radar elevation angle.  This is a
    cumulative effect at each gate along the ray.

    From CSU Radar Meteorology AT 741 Notes

    Parameters
    ----------
    range  : float or array
        Distance to sample volume from radar [m]
    height : float
        Height of the center of radar volume [m]
    elev : float
        Elevation angle [deg]

    Notes
    -----
    This function requires that an array be passed!  If you need just one
       point create a 2 element array with a begin point.
    This is now set up to only accept a 1D array I believe.  May need to
       fix this in the future.
    """
    # Calculate the change in height along the ray
    dh1 = hheight[1:] - hheight[:-1]
    # Add the 0th place in the ray at the beginning
    dh2 = np.insert(dh1, 0, hheight[0])

    # Calculate the change in distance at each gate
    a90r = np.pi/2.  # 90 degrees in radians
    dr = dh2 / (np.tan(a90r - np.deg2rad(elev)))

    # Now calculate the corrected range at each gate
    rnew = np.asarray(range) - np.cumsum(dr)
    return rnew


def beam_block_frac(terrain, beam_height, a):
    """Partial beam blockage fraction. Unitless.

    From Bech et al. (2003), Eqn 2 and Appendix

    Parameters
    ----------
    terrain : float
        Terrain height [m]
    beam_height : float
        Beam height [m]
    a : float
        Half power beam radius [m]

    Notes
    -----
    This procedure uses a simplified interception function where no vertical
      gradient of refractivity is considered.  Other algorithms treat this
      more thoroughly.  However, this is accurate in most cases other than
      the super-refractive case.

    See the the half_power_radius function to calculate variable a

    The heights must be the same units!
    """

    # First find the difference between the terrain and height of
    # radar beam (Bech et al. (2003), Fig.3)
    y = terrain - beam_height

    numer = ((y * np.sqrt(a**2 - y**2)) + (a**2 * np.arcsin(y/a)) +
             (np.pi * a**2 / 2.))

    denom = np.pi * a**2
    return numer / denom

###############
#  Variables  #
###############


def reflectivity(power_t, gain, pulse_width, wavelength, bw_h, bw_v,
                 aloss, rloss, power_return, range, dielectric=0.93):
    """
    Radar reflectivity [mm^6/m^3].

    From Rinehart (1993), Eqn 5.17 (See Eqn 5.14-5.16 also)

    Parameters
    ----------
    power_t : float
        Transmitted power [W]
    gain : float
        Antenna Gain [dB]
    pulse_width : float
        Pulse Width [s]
    wavelength : float
        Radar wavelength [m]
    bwidth : float
        Antenna beamwidth [degrees]
    aloss : float
        Antenna/waveguide/coupler loss [dB]
    rloss : float
        Receiver loss [dB]
    dielectric : float
        Dielectric factor [unitless]
    power_return : float
        Returned power [W]
    range : float or array
        Range to target [m]

    Notes
    -----
    This routine calls the radar_constant function.

    The default is for a dielectric factor value for water.  This can be
    changed by the user, e.g. K=0.208 for particle sizes of equivalent melted
    diameters or K=0.176 for particle sizes of equivalent ice spheres.
    """
    # Call the radar constant function
    C1 = radar_constant(power_t, gain, pulse_width, wavelength, bw_h, bw_v,
                        aloss, rloss)
    return power_return * np.asarray(range)**2 / (C1 * dielectric**2)


def radial_velocity(frequency, wavelength):
    """
    Radial velocity [m/s].

    From Rinehart (1997), Eqn 6.6

    Parameters
    ----------
    frequency : float or array
        Frequency shift [Hz]
    wavelength : float or array
        Radar wavelength [m]

    Notes
    -----
    If arrays are used, either for one okay, or both must be the same length.
    """
    return np.asarray(frequency) * np.asarray(wavelength) / 2.


def cdr(refl_parallel, refl_orthogonal):
    """
    Circular depolarization ratio [dB].

    From Rinehart (1997), Eqn 10.2

    Parameters
    ----------
    refl_parallel : float or array
        Reflectivity in the parallel channel [mm^6/m^3]
    refl_orthogonal : float or array
        Reflectivity in the orthogonal channel [mm^6/m^3]

    Notes
    -----
    Ensure that both powers have the same units!

    Radars that transmit right-hand circular polarization and receive and
       receive both left- and right-hand circular polarization (using two
       antennas) and acquiring the same pulse.
    The parallel (orthogonal) component refers to the same
      (opposite) polarization as transmitted. Non-spericity of hydrometeors
      may be detected (inf long, thin scatterers have CDR = 0 dB, while perfect
     spheres have CDR = -infinity

    Can also use power measurements instead of reflectivity.

    If arrays are used, either for one okay, or both must be the same length.
    """
    return 10. * np.log10(
        np.asarray(refl_parallel)/np.asarray(refl_orthogonal))


def ldr(z_h, z_v):
    """
    Linear depolarization ratio [dB].

    From Rinehart (1997), Eqn 10.3

    Parameters
    ----------
    z_h : float or array
        Horizontal reflectivity [mm^6/m^3]
    z_v : float or array
        Vertical reflectivity [mm^6/m^3]

    Notes
    -----
    Ensure that both powers have the same units!

    Uses both polarizations in a dual-pol radar from a single pulse.

    Perfect spheres yield LDR => -infinity (though antenna limitations limit
    LDR values to -40 dB for small spheres.
    Long, thin targets, LDR => 0

    Typical values in the range -15 > LDR > -35 dB

    If arrays are used, either for one okay, or both must be the same length.
    """
    return 10. * np.log10(np.asarray(z_h) / np.asarray(z_v))


def zdr(z_h, z_v):
    """
    Differential reflectivity [dB].

    From Rinehart (1997), Eqn 10.3 and Seliga and Bringi (1976)

    Parameters
    ----------
    z_h : float or array
        Horizontal reflectivity [mm^6/m^3]
    z_v : float or array
        Vertical reflectivity [mm^6/m^3]

    Notes
    -----
    Ensure that both powers have the same units!

    Alternating horizontally and linearly polarized pulses are averaged.

    Notes
    -----
    If arrays are used, either for one okay, or both must be the same length.
    """
    return 10. * np.log10(np.asarray(z_h) / np.asarray(z_v))


def zdp(z_h, z_v):
    """
    Reflectivity difference [dB].

    From Rinehart (1997), Eqn 10.3

    Parameters
    ----------
    z_h : float
        Horizontal reflectivity [mm^6/m^3]
    z_v : float
        Horizontal reflectivity [mm^6/m^3]

    Notes
    -----
    Ensure that both powers have the same units!

    Alternating horizontally and linearly polarized pulses are averaged.
    """
    zh = np.atleast_1d(z_h)
    zv = np.atleast_1d(z_v)
    if len(zh) != len(zv):
        raise ValueError('Input variables must be same length')
        return

    zdp = np.full_like(zh, np.nan)
    good = np.where(zh > zv)
    zdp[good] = 10. * np.log10(zh[good] - zv[good])
    return zdp


def hdr(dbz_h, zdr):
    """
    Differential reflectivity [dB] hail signature.

    From Aydin et al. (1986), Eqns 4-5

    Parameters
    ----------
    dbz_h : float or array
        Horizontal reflectivity [dBZ]
    zdr : float or array
        Differential reflectivity [dBZ]

    Notes
    -----
    Ensure that both powers have the same units!

    Positive HDR and strong gradients at edges signify ice. The larger HDR,
       the greater likelihood that ice is present.

    Considerations for this equation (see paper for more details):
       1) Standar error of disdrometer data allowed for
       2) Drop oscillation accounted for based on 50% model of
          Seliga et al (1984)
       3) Lower (27) bound chose to provide constant Zh ref level
       4) Upper cutoff of 60 (may be too low)

    Picca and Ryzhkof (2012) mention that this does not take into account
    the hail melting process.  So use at your own risk!
    """
    zdr = np.atleast_1d(zdr)
    # Set the f(zdr) based upon observations
    f = np.full_like(zdr, np.nan)
    negind = np.where(zdr <= 0)
    lowind = np.where((zdr > 0) & (zdr <= 1.74))
    highind = np.where(zdr > 1.74)
    f[negind] = 27.
    f[lowind] = 19. * zdr[lowind] + 27.
    f[highind] = 60.
    # Calculate HDR
    return np.asarray(dbz_h) - f

#################
#  Attenuation  #
#################


def abs_coeff(diam, wavelength, m):
    """
    Absorption coefficient of a spherical particle. Unitless.

    Doviak and Zrnic (1993), Eqn 3.14a or Battan (1973), Eqn 6.6

    Parameters
    ----------
    diam : float or array
        Particle diameter [m]
    wavelength : float
        Radar wavelength [m]
    m : float
        Complex refractive index [unitless], in the form 7.14 - 2.89j

    Notes
    -----
    An example from Battan (1973) is for water at 0C m=7.14-2.89j for a
       wavelength of 3.21 cm and for ice m=1.78-0.0024j for
       wavelength range from 1-10 cm.
    See Battan (1973) Ch.4 , Tables 4.1 and 4.2 for values from
       Gunn and East (1954).
    Also see Doviak and Zrnic (1993), Fig. 3.3 caption.
    """
    Km_im = np.imag(-1 * k_complex(m))
    Qa = (np.pi**2 * np.asarray(diam)**3 / wavelength) * Km_im
    return Qa


def scat_coeff(diam, wavelength, m):
    """
    Scattering coefficient of a spherical particle. Unitless.

    Doviak and Zrnic (1993), Eqn 3.14b or Battan (1973), Eqn 6.4

    Parameters
    ----------
    diam: float or array
        Particle diameter [m]
    wavelength: float
        Radar wavelength [m]
    m: float
        Complex refractive index [unitless]
    """
    Km_abs = np.absolute(k_complex(m))
    Qs = 2 * np.pi**5 * np.asarray(diam)**6 / (3 * wavelength**4) * (Km_abs**2)
    return Qs


def ext_coeff(diam, wavelength, m):
    """
    Extinction coefficient of a spherical particle. Unitless.

    Doviak and Zrnic (1993), Eqn 3.14b or Battan (1973), Eqn 6.5

    Parameters
    ----------
    diam : float or array
        Particle diameter [m]
    wavelength : float
        Radar wavelength [m]
    m : float
        Complex refractive index [unitless]

    Notes
    -----
    An example from Battan (1973) is for water at 0C m=7.14-2.89j for a
       wavelength of 3.21 cm and for ice m=1.78-0.0024j for
       wavelength range from 1-10 cm.
    See Battan (1973) Ch.4 , Tables 4.1 and 4.2 for values from
       Gunn and East (1954).
    Also see Doviak and Zrnic (1993), Fig. 3.3 caption.
    """
    Qa = abs_coeff(np.asarray(diam), wavelength, m)
    Qs = scat_coeff(np.asarray(diam), wavelength, m)
    Qe = Qa + Qs
    return Qe


def k_complex(m):
    """
    Parameters
    ----------
    m : float
        Complex refractive index [unitless]
    """
    return (m**2 - 1) / (m**2 + 2)



"""
csu_t_traps.py

Brenda Dolan, CSU, October 2018
bdolan@atmos.colostate.edu

Creating Trapazoidal functions for the temperature fuzzy logic.
In this sense we can penalize certain temperatures for certain HID types, and 
therefore reduce some of the weighting on temperature while also reducing the harsh
boundaries between some of the species that are directly related to temperature.
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
import numpy as np


def get_hid_traps(c,x_arr):
    if c == 0 :
        #Drizzle
        x1 = -3
        x2 = -1
        beta = t_trap_dz(x_arr,x1,x2)
    elif c==1:
        #Rain
        x1 = -10
        x2 = -2
        beta = t_trap_rain(x_arr,x1,x2)
    elif c==2:
        #Ice Crystals
        x1 = 0
        x2 = 5
        beta = t_trap_ic(x_arr,x1,x2)
    elif c ==3:
        #Aggregates
        x1 = -35
        x2 = -25
        x3 = 0
        x4 = 5
        beta = t_trap_agg(x_arr,x1,x2,x3,x4)
    elif c==4:
        #Wet Snow
        x1 = -3
        x2 = 3
        beta = t_trap_ws(x_arr,x1,x2)
    elif c==5:
        #Vertical ICe. Should exactly match Ice Crystals
        x1 = 0
        x2 = 5
        beta = t_trap_ic(x_arr,x1,x2)
    elif c ==6:
        #Low Density Graupel
        x1 = -40
        x2 = -30
        x3 = 0
        x4 = 5
        beta = t_trap_ldg(x_arr,x1,x2,x3,x4)
    elif c == 7:
        #High Density Graupel
        x1 = -30
        x2 = -20
        x3 = 5
        x4 = 30
        beta = t_trap_hdg(x_arr,x1,x2,x3,x4)
    elif c == 8:
        #Hail
        x1 = -30
        x2 = -20
        x3 = 5
        x4 = 15
        beta = t_trap_ha(x_arr,x1,x2,x3,x4)
    elif c == 9:
        #Big Drops
        x1 = -2
        x2 = 1
        beta = t_trap_bd(x_arr,x1,x2)
    else:
        print('HID type is out of range. Returning')
        return
    return beta

def t_trap_agg(x_arr,x1,x2,x3,x4):
    beta = np.zeros_like(x_arr)
    wh1 = np.where(x_arr <= x1)
    beta[wh1] = 0.5
    wh2 = np.where(np.logical_and(x_arr>=x2,x_arr<=x3))
    beta[wh2] = 1
    wh3 = np.where(x_arr >= x4)
    beta[wh3] = -1
    
    slope_x1_x2 = (.5)/np.abs(np.abs(x1)-np.abs(x2))
    slope_x3_x4 = (1.)/np.abs(np.abs(x4)-np.abs(x3))
    wh4 = np.where(np.logical_and(x_arr >= x1,x_arr<=x2))
    test1 = slope_x1_x2*(np.abs(x1)+x_arr)+0.5
    beta[wh4] = test1[wh4]
    wh5 = np.where(np.logical_and(x_arr >=x3,x_arr<=x4))
    test2 = slope_x3_x4*(np.abs(x4)-x_arr)
    beta[wh5] =test2[wh5]
    return beta

def t_trap_bd(x_arr,x1,x2):
    beta = np.zeros_like(x_arr)
    wh1 = np.where(x_arr <= x1)
    beta[wh1] = -1
    wh2 = np.where(x_arr >=x2)
    beta[wh2] = 1
    
    slope_x1_x2 = (2.)/np.abs(np.abs(x1)+np.abs(x2))
#    print(slope_x1_x2)
    wh4 = np.where(np.logical_and(x_arr >= x1,x_arr<=x2))
    test1 = np.abs(slope_x1_x2*(x_arr-x1)/(x2-x1))
    beta[wh4] = test1[wh4]
    return beta

def t_trap_dz(x_arr,x1,x2):
    beta = np.zeros_like(x_arr)
    wh1 = np.where(x_arr <= x1)
    beta[wh1] = -1
    wh2 = np.where(x_arr >=x2)
    beta[wh2] = 1
    
    slope_x1_x2 = (2.)/np.abs(np.abs(x1)+np.abs(x2))
    print(slope_x1_x2)
    wh4 = np.where(np.logical_and(x_arr >= x1,x_arr<=x2))
    test1 = np.abs(slope_x1_x2*(x_arr-x1)/(x2-x1))
    beta[wh4] = test1[wh4]
    return beta
    
def t_trap_ha(x_arr,x1,x2,x3,x4):
    beta = np.zeros_like(x_arr)
    wh1 = np.where(x_arr <= x1)
    beta[wh1] = 0.2
    wh2 = np.where(np.logical_and(x_arr>=x2,x_arr<=x3))
    beta[wh2] = 0.8
    wh3 = np.where(x_arr >= x4)
    beta[wh3] = 0.2
    
    slope_x1_x2 = (0.6)/np.abs(np.abs(x1)-np.abs(x2))
    slope_x3_x4 = (0.6)/np.abs(np.abs(x4)-np.abs(x3))
    wh4 = np.where(np.logical_and(x_arr >= x1,x_arr<=x2))
    test1 = slope_x1_x2*(np.abs(x1)+x_arr)+0.2
    beta[wh4] = test1[wh4]
    wh5 = np.where(np.logical_and(x_arr >=x3,x_arr<=x4))
    test2 = slope_x3_x4*(np.abs(x4)-x_arr)+0.2
    beta[wh5] =test2[wh5]
    return beta    
    
def t_trap_hdg(x_arr,x1,x2,x3,x4):
    beta = np.zeros_like(x_arr)
    wh1 = np.where(x_arr <= x1)
    beta[wh1] = 0.3
    wh2 = np.where(np.logical_and(x_arr>=x2,x_arr<=x3))
    beta[wh2] = 1
    wh3 = np.where(x_arr >= x4)
    beta[wh3] = -1
    
    slope_x1_x2 = (0.7)/np.abs(np.abs(x1)-np.abs(x2))
    slope_x3_x4 = (2)/np.abs(np.abs(x4)-np.abs(x3))
    wh4 = np.where(np.logical_and(x_arr >= x1,x_arr<=x2))
    test1 = slope_x1_x2*(np.abs(x1)+x_arr)+0.3
    beta[wh4] = test1[wh4]
    wh5 = np.where(np.logical_and(x_arr >=x3,x_arr<=x4))
    test2 = slope_x3_x4*(np.abs(x4)-x_arr)-1
    beta[wh5] =test2[wh5]
    return beta 
    
    
def t_trap_ic(x_arr,x1,x2):
    beta = np.zeros_like(x_arr)
    wh1 = np.where(x_arr <= x1)
    beta[wh1] = 1
    wh3 = np.where(x_arr >= x2)
    beta[wh3] = -1
    
    slope_x1_x2 = (1.)/np.abs(x2+x1)
    wh4 = np.where(np.logical_and(x_arr >= x1,x_arr<=x2))
    test1 = 1-np.abs(slope_x1_x2*(x_arr-x1))
    beta[wh4] = test1[wh4]

    return beta           
    
def t_trap_ldg(x_arr,x1,x2,x3,x4):
    beta = np.zeros_like(x_arr)
    wh1 = np.where(x_arr <= x1)
    beta[wh1] = 0.5
    wh2 = np.where(np.logical_and(x_arr>=x2,x_arr<=x3))
    beta[wh2] = 1
    wh3 = np.where(x_arr >= x4)
    beta[wh3] = -1
    
    slope_x1_x2 = (0.5)/np.abs(np.abs(x1)-np.abs(x2))
    slope_x3_x4 = (1)/np.abs(np.abs(x4)-np.abs(x3))
    wh4 = np.where(np.logical_and(x_arr >= x1,x_arr<=x2))
    test1 = slope_x1_x2*(np.abs(x1)+x_arr)+0.5
    beta[wh4] = test1[wh4]
    wh5 = np.where(np.logical_and(x_arr >=x3,x_arr<=x4))
    test2 = slope_x3_x4*(np.abs(x4)-x_arr)
    beta[wh5] =test2[wh5]
    return beta 
   
def t_trap_rain(x_arr,x1,x2):
    beta = np.zeros_like(x_arr)
    wh1 = np.where(x_arr <= x1)
    beta[wh1] = -1
    wh3 = np.where(x_arr >= x2)
    beta[wh3] = 1
    
    slope_x1_x2 = (1.)/np.abs(x2+x1)
    wh4 = np.where(np.logical_and(x_arr >= x1,x_arr<=x2))
    test1 = np.abs(slope_x1_x2*(x_arr-x1))
    beta[wh4] = test1[wh4]

    return beta           

def t_trap_ws(x_arr,x1,x2):
    beta = np.zeros_like(x_arr)
    wh1 = np.where(x_arr <= x1)
    beta[wh1] = -1
    wh3 = np.where(x_arr >= x2)
    beta[wh3] = -1
    wh2 = np.where(np.logical_and(x_arr>=x1,x_arr<=x2))
    beta[wh2] = 1
    return beta 



# -*- coding: utf-8 -*-
"""
Functions to estimate specific differential phase (KDP) using various methods.
"""

from __future__ import division, print_function
import numpy as np
from numpy import linalg
from scipy.signal import firwin
from warnings import warn
from .calc_kdp_ray_fir import calc_kdp_ray_fir
# import time

VERSION = '1.6'

# Used by FIR coefficient function (get_fir)
FIR_GS = 150.0
FIR_WIN = 3.0
FIR_ORDER = 20
FIR_GAIN = 1.0
FIR_FREQ = 0.08
FIR_STD = 28.0
KM2M = 1000.0
STD_GATE = 11


def calc_kdp_bringi(dp=None, dz=None, rng=None, thsd=12, nfilter=1,
                    bad=-32768, gs=FIR_GS, window=FIR_WIN, std_gate=STD_GATE):
    """
    Overview
    --------
    This is an old algorithm that uses an FIR filter to process differential
    phase and extract specific differential phase. It works on polarimetric
    radar data. It is based on code provided by V. N. Bringi and Yanting Wang
    of CSU Electrical Engineering. It assumes differential phase has been
    unfolded already. You can send this function either 1D or 2D arrays of
    data. If 2D, it assumes the first index is azimuth so it will loop over
    that, calculating KDP along individual rays.

    Steps
    -----
    1. Standard deviation of differential phase is calculated and used to
       QC the phase data. The stdev calculation uses up to std_gate consecutive
       gates regardless of gate spacing.
    2. Differential phase is filtered using the FIR filter, which has been
       tuned to the number of gates contained within the FIR window. This
       algorithm only works for window / gate spacing = even number.
    3. Specific differential phase is calculated by consulting reflectivity.
       As reflectivity declines progressively more and more gates are needed
       in the window used to fit a line to the filtered phase. Specific
       differential phase is half the slope of that line.

    Reference
    ---------
    Timothy J. Lang, David A. Ahijevych, Stephen W. Nesbitt, Richard E.
    Carbone, Steven A. Rutledge, and Robert Cifelli, 2007: Radar-Observed
    Characteristics of Precipitating Systems during NAME 2004. J. Climate,
    20, 1713–1733. doi: https://dx.doi.org/10.1175/JCLI4082.1

    Arguments
    ---------
    dp = Differential phase (deg, 1D or 2D array)
    dz = Reflectivity (dBZ, 1D or 2D array)
    rng = Range (km, 1D or 2D array -
          use np.meshgrid() first tp make rng 2D if needed)
    thsd = Threshold for standard deviation of differential phase, above which
           the data are not considered when filtering or calculating specific
           differential phase. The user can specify a 1D vector of spatially
           varying thresholds instead (i.e., vary by range).
    nfilter = Number of times to apply the FIR filter
    bad = Value for bad/missing data
    gs = Gate spacing of radar (meters)
    window = Changes window over which FIR filter is applied (km). Also affects
             the width of the adaptive KDP calculations.
    std_gate = Number of gates for standard deviation of phase calculation.
               Must be odd or function will just set it to the default value.

    Returns
    -------
    kd_lin = Specific differential phase (deg/km, 1D or 2D array)
    dp_lin = Filtered differential phase (deg, 1D or 2D array)
    sd_lin = Standard deviation of diff. phase (deg, 1D or 2D array)

    """
    # Quick check on all vars. Used keywords so order doesn't matter.
    if dp is None or dz is None or rng is None:
        warn('Missing needed variables (dp, dz, and/or rng), failing ...')
        return
    if np.ndim(dp) != np.ndim(dz) or np.ndim(dp) != np.ndim(rng):
        warn('Array sizes don\'t match, failing ...')
        return
    if std_gate % 2 != 1:
        warn('std_gate must be odd, using ' + str(STD_GATE) +
             ' gates as the default window')
        std_gate = STD_GATE
    fir = get_fir(gs=gs, window=window)
    if fir is None:
        print('Fix window/gs to be even, failing ...')
        return None, None, None

    # Following lines ensure right dtype passed to Cython extensions (if used)
    dp = np.array(dp).astype('float32')
    dz = np.array(dz).astype('float32')
    rng = np.array(rng).astype('float32')
    fir['coef'] = np.array(fir['coef']).astype('float64')

    if not hasattr(thsd, '__len__'):
        thsd = np.zeros_like(dp) + thsd
    # If array is 2D, then it assumes the first index refers to azimuth.
    # Thus it loops over that.
    if np.ndim(dp) == 2:
        kd_lin = np.zeros_like(dp) + bad
        dp_lin = np.zeros_like(dp) + bad
        sd_lin = np.zeros_like(dp) + 100.0
        for ray in np.arange(np.shape(dp)[0]):
            dpl = len(dp[ray])
            kd_lin[ray], dp_lin[ray], sd_lin[ray] = calc_kdp_ray_fir(
                dpl, dp[ray], dz[ray], rng[ray], thsd[ray],
                nfilter, bad, fir['order'], fir['gain'], fir['coef'], std_gate)
#             kd_lin[ray], dp_lin[ray], sd_lin[ray] = \
#                 _calc_kdp_ray(dp[ray], dz[ray], rng[ray], thsd=thsd,
#                               nfilter=nfilter, bad=bad, fir=fir)
    # Or
    elif np.ndim(dp) == 1:
        kd_lin, dp_lin, sd_lin = calc_kdp_ray_fir(
            len(dp), dp, dz, rng, thsd, nfilter, bad,
            fir['order'], fir['gain'], fir['coef'], std_gate)
#        kd_lin, dp_lin, sd_lin = _calc_kdp_ray(
#             dp, dz, rng, thsd=thsd, fir=fir, nfilter=nfilter, bad=bad)
    else:
        warn('Need 2D or 1D array, failing ...')
        return
    return kd_lin, dp_lin, sd_lin


def get_fir(gs=FIR_GS, window=FIR_WIN):
    """
    gs = Gate Spacing (m)
    window = Filter Window (km)
    window divided by gs should be an even number!
    """
    fir = {}
    fir['order'] = np.int32(window * KM2M / gs)
    if fir['order'] % 2 != 0:
        warn('gs / window must be an even number! #Failing ...')
        return
    fir['gain'] = FIR_GAIN
    # ratio = FIR_GS / gs
    ratio = fir['order'] / FIR_ORDER
    freq = FIR_FREQ / ratio
    std = ratio * FIR_STD
    fir['coef'] = firwin(fir['order'] + 1, freq, window=('gaussian', std))
    # print('debug', fir)
    return fir


def _calc_kdp_ray(dp, dz, rng, thsd=12, nfilter=1, bad=-32768, fir=None):
    """
    Pure Python approach to filtering phase and estimating KDP. Currently
    disabled due to performance issues.

    Arguments
    ---------
    dp = 1D ray of differential phase
    dz = 1D ray of reflectivity
    rng = 1D ray of range
    thsd = Scalar or 1D ray of diff phase standard deviation thresholds
    nfilter = Number of times to filter the data
    bad = Bad/missing data value
    fir = Dictionary containing FIR filter parameters

    Returns
    -------
    kd_lin = Specific differential phase (deg/km, 1D array)
    dp_lin = Filtered differential phase (deg, 1D array)
    sd_lin = Standard deviation of diff. phase (deg, 1D array)
    """
    # Define needed variables
    kd_lin = np.zeros_like(rng) + bad
    sd_lin = np.zeros_like(rng) + 100.0
    # User can provide a spatially varying stddev(dp) threshold
    if not hasattr(thsd, '__len__'):
        thsd = np.zeros_like(rng) + thsd
    length = len(rng)
    lin = np.arange(length)
    # Half window size for calculating stdev of phase (fixed @ 11 gates)
    half_std_win = 5
    half_fir_win = fir['order'] // 2  # Half window size for FIR filtering
    y = np.zeros(length) + bad  # Dummy variable to store filtered phase
    z = 1.0 * dp  # Dummy variable to store un/pre-processed phase
    # print(time.time() - begin_time, 'seconds since start (DEF)')

    #####################################################################
    # Calculate standard deviation of phidp
    mask = dp != bad
    for i in lin[mask]:
        index1 = np.int32(i - half_std_win)
        index2 = np.int32(i + half_std_win)
        if index1 >= 0 and index2 < length - 1:
            yy = dp[index1:index2]
            tmp_mask = mask[index1:index2]
            if len(yy[tmp_mask]) > half_std_win:
                sd_lin[i] = _quick_std(yy, tmp_mask)

    # ------------- MAIN LOOP of Phidp Adaptive Filtering ------------------
    # FIR FILTER SECTION
    for mloop in np.arange(nfilter):
        mask = np.logical_and(sd_lin <= thsd, z != bad)
        for i in lin[mask]:
            index1 = np.int32(i - half_fir_win)
            index2 = np.int32(i + half_fir_win)
            if index1 >= 0 and index2 < length - 1:
                yy = z[index1:index2+1]
                xx = rng[index1:index2+1]
                tmp_mask = mask[index1:index2+1]
                siz = len(yy[tmp_mask])
                if siz > 0.8 * fir['order']:
                    if siz < fir['order'] + 1:
                        result = _leastsqrs(xx, yy, siz, tmp_mask)
                        yy[~tmp_mask] = result[0] * xx[~tmp_mask] + result[1]
                    y[i] = fir['gain'] * np.dot(fir['coef'], yy)
        z = 1.0 * y  # Enables re-filtering of processed phase
    dp_lin = 1.0 * z
    # print(time.time() - begin_time, 'seconds since start (FDP)')
    # *****************END LOOP for Phidp Adaptive Filtering******************

    # CALCULATE KDP
    # Default value for nadp is half_fir_win, but varies based on Zh
    nadp = np.int16(0 * dz + half_fir_win)
    tmp_mask = dz < 35
    nadp[tmp_mask] = 3 * half_fir_win
    tmp_mask = np.logical_and(dz >= 35, dz < 45)
    nadp[tmp_mask] = 2 * half_fir_win
    mask = dp_lin != bad
    for i in lin[mask]:
        index1, index2 = _get_nadp_indices(nadp, i)
        if index1 >= 0 and index2 <= length:
            tmp_mask = mask[index1:index2]
            xx = rng[index1:index2]
            siz = len(xx[tmp_mask])
            # Improved Kdp based on LSE fit to Adap filt Phidp
            if siz >= 0.8 * nadp[i]:
                yy = dp_lin[index1:index2]
                kd_lin[i] = _fit_line_and_get_kdp(xx, yy, siz, tmp_mask)
    # *******************END KDP CALCULATION****************************
    # print(time.time() - begin_time, 'seconds since start (KDP/Done)')
    return kd_lin, dp_lin, sd_lin


def _leastsqrs(xx, yy, siz, tmp_mask):
    """
    Following is faster than np.polyfit
    e.g., return np.polyfit(xx[tmp_mask], yy[tmp_mask], 1)
    """
    A = np.array([xx[tmp_mask], np.ones(siz)])
    return linalg.lstsq(A.T, yy[tmp_mask])[0]


def _get_nadp_indices(nadp, i):
    half_nadp = nadp[i] / 2
    return np.int32(i - half_nadp), np.int32(i + half_nadp + 1)


def _fit_line_and_get_kdp(xx, yy, siz, tmp_mask):
    result = _leastsqrs(xx, yy, siz, tmp_mask)
    return 0.5 * result[0]


def _quick_std(array, mask):
    """Following is faster than np.std()"""
    a = array[mask]
    m = a.mean()
    c = a - m
    return (np.dot(c, c) / a.size)**0.5



# cython: boundscheck=False
# cython: language_level=2

import cython
from libc.math cimport sqrt
cimport numpy as np
import numpy as np


def LSE(double[:] x, double[:] y, int n, float bad):
    """
    This is a Least Square Estimate subroutine to fit a linear
    equation for (xi, yi) (i = 1, ..., n), so that yi = a * xi + b
    INPUTs: x(i), y(i), n, (i = 1, ..., n ).
    OUTPUTs: a (slope), b (intercept).
    Li Liu - Sep. 23, 92
    """
    cdef:
        float a
        float b
        float xsum = 0.0
        float ysum = 0.0
        float xxsum = 0.0
        float xysum = 0.0
        float det
        float total
        Py_ssize_t i

    total = float(n)
    for i in range(n):
        if x[i] == bad or y[i] == bad:
            total -= 1.0
        else:
            xsum += x[i]
            ysum += y[i]
            xxsum += x[i] * x[i]
            xysum += x[i] * y[i]

    det = total * xxsum - xsum**2
    a = (total * xysum - xsum * ysum) / det
    b = (ysum * xxsum - xsum * xysum) / det
    return a, b


def calc_kdp_ray_fir(
        int ngates, float[:] dp, float[:] dz, float[:] rng,
        float[:] thsd, int nf, float bad, int fir_order, float fir_gain,
        double[:] fir_coeff, int std_gate):
    """
    Calculate KDP along a 1-D ray of radar data, using an FIR filter to first
    smooth the differential phase. Standard deviation of phase is also
    calculated.

    Arguments
    ---------
    dp = 1D ray of differential phase
    dz = 1D ray of reflectivity
    rng = 1D ray of range
    thsd = Scalar or 1D ray of diff phase stddev thresholds
    nf = Number of times to filter the data
    bad = Bad/missing data value
    fir_order = FIR filter order
    fir_gain = FIR filter gain
    fir_coeff = FIR filter coefficients
    std_gate = Number of gates to use for diff phase stddev calc

    Returns
    -------
    kd_lin = Specific differential phase (deg/km, 1D array)
    dp_lin = Filtered differential phase (deg, 1D array)
    sd_lin = Standard deviation of diff. phase (deg, 1D array)
    """
    # Define needed variables
    cdef:
        np.ndarray[np.float64_t, ndim=1] xx = bad + np.zeros(500, dtype=np.float64)
        np.ndarray[np.float64_t, ndim=1] yy = bad + np.zeros(500, dtype=np.float64)
        np.ndarray[np.float32_t, ndim=1] y = np.zeros(ngates, dtype=np.float32)
        np.ndarray[np.float32_t, ndim=1] z = np.zeros(ngates, dtype=np.float32)
        np.ndarray[np.float32_t, ndim=1] kd_lin = np.zeros(ngates, dtype=np.float32)
        np.ndarray[np.float32_t, ndim=1] dp_lin = np.zeros(ngates, dtype=np.float32)
        np.ndarray[np.float32_t, ndim=1] sd_lin = np.zeros(ngates, dtype=np.float32)
        int half_std_win, nadp, half_fir_win
        int index1, index2, N, half_nadp
        double X, A, V, W, aa, bb
        Py_ssize_t i, j, mloop

    # Half window size for calculating stdev phase
    half_std_win = (std_gate - 1) / 2
    # Half window size for FIR filtering
    half_fir_win = fir_order / 2

    #####################################################################
    # Calculate standard deviation of phidp
    for i in range(ngates):
        kd_lin[i] = bad
        sd_lin[i] = 100.0
        y[i] = bad
        z[i] = dp[i]
        index1 = i - half_std_win
        index2 = i + half_std_win
        if index1 >= 0 and index2 <= ngates - 1:
            N = 0
            A = 0.0
            V = 0.0
            for j in range(index1, index2 + 1, 1):
                if dp[j] != bad:
                    # Standard deviation algorithm
                    X = dp[j]
                    if N <= 0:
                        W = X
                    N += 1
                    D = X - W
                    V = (N - 1) * (D - A)**2 / N + V
                    A = (D - A) / N + A
            if N > half_std_win:
                sd_lin[i] = sqrt(V / N)

    # ------------- MAIN LOOP of Phidp Adaptive Filtering ------------------
    # FIR FILTER SECTION
    for mloop in range(nf):
        for i in range(half_fir_win, ngates - half_fir_win, 1):
            if sd_lin[i] <= thsd[i] and z[i] != bad:
                index1 = i - half_fir_win
                index2 = i + half_fir_win
                N = 0
                for j in range(index1, index2 + 1, 1):
                    if sd_lin[j] <= thsd[j] and z[j] != bad:
                        yy[N] = z[j]
                        xx[N] = rng[j]
                        N += 1

                # Now fill in gaps if they aren't too big
                if float(N) > (0.8 * float(fir_order)):
                    if N < fir_order + 1:
                        aa, bb = LSE(xx, yy, N, bad)
                        for j in range(index1, index2 + 1, 1):
                            if z[j] == bad:
                                z[j] = aa * rng[j] + bb

                    # Now do the FIR filtering
                    A = 0.0
                    for j in range(index1, index2 + 1, 1):
                        A += fir_coeff[j - index1 + 1] * z[j]
                    y[i] = A * fir_gain
        for i in range(ngates):
            z[i] = y[i]  # Enables re-filtering of processed phase
    for i in range(ngates):
        dp_lin[i] = z[i]

    # *****************END LOOP for Phidp Adaptive Filtering******************

    # CALCULATE KDP
    # Default value for nadp is half_fir_win, but varies based on Zh
    for i in range(ngates):
        if dz[i] != bad:
            if dz[i] >= 45.0:
                nadp = half_fir_win
            if dz[i] >= 35.0 and dz[i] < 45.0:
                nadp = 2 * half_fir_win
            if dz[i] < 35.0:
                nadp = 3 * half_fir_win
            half_nadp = nadp / 2
            index1 = i - half_nadp
            index2 = i + half_nadp
            N = 0
            for j in range(index1, index2, 1):
                if index1 >= 0 and index2 <= ngates - 1:
                    if dp_lin[j] != bad:
                        yy[N] = dp_lin[j]
                        xx[N] = rng[j]
                        N += 1
            if float(N) > (0.8 * float(nadp)):
                aa, bb = LSE(xx, yy, N, bad)
                kd_lin[i] = 0.5 * aa

    # *******************END KDP CALCULATION****************************
    return kd_lin, dp_lin, sd_lin


def hid_beta_f(int ngates, float[:] x_arr, float a, float b, float m):
    """
    Beta function calculator
    ngates = Number of gates
    x_arr = Array to process
    a, b , m = Beta function parameters
    """

    cdef:
        np.ndarray[np.float32_t, ndim=1] beta = np.zeros(ngates, dtype=np.float32)
        Py_ssize_t i

    for i in range(ngates):
        beta[i] = 1.0 / (1.0 + (((x_arr[i] - m) / a)**2.0)**b)
        # beta[i] = 1.0 / (1.0 + np.power(np.power((x_arr[i] - m) / a, 2.0), b))

    return beta



# -*- coding: utf-8 -*-
"""
Python remake of Brenda Dolan's code to calculate water and ice mass.

This program will calculate ice and water mass at specified height levels
given Z and ZDR at each grid point.

Based on Code from A. Rowe and T. Lang from L. Nelson and L. Carey.

Brenda Dolan
February 27, 2012
bdolan@atmos.colostate.edu
"""
from __future__ import division
import numpy as np
import warnings

DEFAULT_HFRZ = 4.5  # km MSL


def calc_liquid_ice_mass(dbz, zdr, z, T=None, Hfrz=None, method='cifelli',
                         fit_a=None, fit_b=None):
    """
    This function uses the ZDP method to discriminate between

    Arguments:

    Must be same shape
    ------------------
    dbz = Reflectivity (dBZ)
    zdr = Differential Reflectivity (dB)
    z = Height (km MSL)
    T = Temperature (deg C)

    Other keywords
    --------------
    method = Name of method to determine fit_a and fit_b for rain line.
             Options are 'cifelli' and 'cr1995' which correspond to
             Cifelli et al. 2002 and Carey and Rutledge (1995). If you put
             anything else here, it will use Carey and Rutledge (2000).
    fit_a, fit_b = Fit coefficients to the rain line. If user doesn't specify,
                   then the reference specified in method will be used.

    Scalar only
    -----------
    Hfrz = Height of freezing level (km MSL), if known; will be calculated
           from sounding provided otherwise.

    All need to be same array or scalar structure. If T == None, then
    function will assume default arrangement of delta_thresh.

    No error checking done, user is responsible for their own bad data masking.

    """
    len_flag = hasattr(dbz, '__len__')
    if len_flag:
        shape = np.shape(dbz)
        dbz = dbz.ravel()
        zdr = zdr.ravel()
        z = z.ravel()
        if T is not None:
            T = T.ravel()
    else:
        dbz = np.array([dbz])
        zdr = np.array([zdr])
        z = np.array([z])
        if T is not None:
            T = np.array([T])
    Hfrz = _get_hfrz(T, z, Hfrz)
    if Hfrz is None:
        return
    delta_thresh = get_delta_thresh(z, Hfrz)
    mass_w, mass_i = _liquid_ice_calcs(dbz, zdr, delta_thresh, fit_a=fit_a,
                                       fit_b=fit_b)
    if len_flag:
        return np.reshape(mass_w, shape), np.reshape(mass_i, shape)
    else:
        return mass_w[0], mass_i[0]


def linearize(dz):
    """dz = Reflectivity (dBZ), returns Z (mm^6 m^-3)"""
    return 10.0**(dz / 10.0)


def get_freezing_altitude(T, z):
    """
    T = temperature (deg C), z = altitude (any units)
    Assumes T & z are aligned, unmasked ndarrays of same shape.
    Returns highest possible freezing altitude.
    Partial soundings (i.e., no pass thru freezing altitude) get fitted to
    a regression line and Hfrz is then estimated via the intercept.
    """
    T = T.ravel()
    z = z.ravel()
    try:
        # Maybe there's already one or more zero points in sounding
        return np.max(z[T == 0])
    except:
        # Following gets indices of sorted z array, for comparison w/ T.
        # Commented two lines below were slower than necessary.
        # zs = sorted(z)
        # Tz = T[np.argsort(z)]
        sin = np.argsort(z)
        zs = z[sin]
        Tz = T[sin]

        # If sounding passes thru 0, then look immediately above and below
        if np.max(T) > 0 and np.min(T) < 0:
            zarg = np.argmax(zs[Tz > 0])
            T2 = Tz[zarg:zarg+2]
            z2 = zs[zarg:zarg+2]
            if T2[1] < T2[0]:
                return np.interp(0.0, T2[::-1], z2[::-1])
            else:
                return np.interp(0.0, T2, z2)
        # Sounding doesn't pass thru 0, trying linear regression
        elif np.max(T) < 0 or np.min(T) > 0:
            reg = np.polyfit(Tz, zs, 1)
            return reg[1]


def get_delta_thresh(z, Hfrz):
    """
    Makes assumptions about change in T with height,
    so an abnormal T profile may lead to inaccurate results.
    TJL - Currently feel set too aggressively, leading to turning obvious
    hail gates (high Z, low ZDR) into all-rain, giving huge LWCs.
    """
    delta_thresh = np.zeros_like(z) + 1.1
    cond = z <= Hfrz
    delta_thresh[cond] = (-8.0/3.0) * (z[cond] - Hfrz) + 2.0
    return delta_thresh


def get_linear_fits(method='cifelli'):
    """
    Get coefficients from linear fits to Zdp and Zh
    method = str, Reference for method
        'cifelli': Cifelli et al. (2002) - Amazon
        'cr1995': Carey and Rutledge (1995) - Colorado
        Default: Carey and Rutledge (2000) - Darwin
    """
    # TJL - Does Carey and Rutledge (1995) have another rain line?
    if method == 'cifelli':
        # These are Cifelli et al. (2002) for Amazon
        fit_a = 0.8178
        fit_b = 12.5088
    elif method == 'cr1995':
        # These are Carey and Rutledge (1995) for Colorado (?)
        fit_a = 0.9091
        fit_b = 8.5091
    else:
        # These are Carey and Rutledge (2000) for MCTEX
        fit_a = 0.77
        fit_b = 14.0
    return fit_a, fit_b


def calc_zh_zv(dbz, zdr):
    Zh = linearize(dbz)
    return Zh, Zh / linearize(zdr)  # Vertical component of reflectivity


def calc_zdp(Zh, Zv):
    Z_dp = Zh - Zv  # Difference reflectivity
    Z_dp[Z_dp <= 0] = 0.000000000001
    return 10.0 * np.log10(Z_dp)  # Convert to decibel units


def _get_hfrz(T, z, Hfrz):
    if T is None:
        if Hfrz is None:
            Hfrz = DEFAULT_HFRZ
    else:
        if Hfrz is None:
            if hasattr(T, '__len__'):
                Hfrz = get_freezing_altitude(T, z)
            else:
                warnings.warn('Given a scalar for T and no Hfrz, failing')
    return Hfrz


def _liquid_ice_calcs(dbz, zdr, delta_thresh, fit_a=None, fit_b=None):
    # Set up arrays to hold the output data
    mass_w = dbz * 0.0
    mass_i = dbz * 0.0

    # Get Zdp
    Zh, Zv = calc_zh_zv(dbz, zdr)
    Zdp = calc_zdp(Zh, Zv)

    # Determine ice fraction and partition reflectivity
    if fit_a is None:
        fit_a, fit_b = get_linear_fits(method='cifelli')
    dbz_est = fit_a * Zdp + fit_b  # Estimate the best fit line
    delta_z = dbz - dbz_est
    delta_z[delta_z < 0] = 0.0
    ice_frac = 1.0 - 10.0**(-0.1 * delta_z)
    Zh_rain = Zh * (1.0 - ice_frac)
    Zh_ice = Zh - Zh_rain

    # Check against delta_thresh
    check = delta_z - delta_thresh
    above = check >= 0
    below = check < 0

    # Z-M relationships for water and ice (Cifelli et al. 2002)
    mass_w[below] = 0.70 * 10.0**(-3) * Zh[below]**(0.886) * \
        (Zh[below] / Zv[below])**(-4.159)
    # mass_i[below] = 0.0 anyway
    mass_w[above] = 3.44 * 10.0**(-3) * Zh_rain[above]**(4.0/7.0)
    mass_i[above] = 1000.0 * np.pi * 917.0 * (4.0 * 10.0**6)**(3.0/7.0) * \
        ((5.28 * 10.0**(-18) * Zh_ice[above]) / 720.0)**(4.0/7.0)
    mass_w[zdr < 0] = 0.0  # Negative ZDR = all ice
    return mass_w, mass_i  # Units = g m^-3



# -*- coding: utf-8 -*-
"""
Package that contains a variety of functions for polarimetric radar retrievals.

=========================================
Top-level package (:mod:`csu_radartools`)
=========================================

.. currentmodule:: csu_radartools

"""

from ._version import __version__

__all__ = [s for s in dir() if not s.startswith('_')]



# -*- coding: utf-8 -*-
"""
Python remake of Brenda Dolan's code to calculate water and ice mass.

Brody Fuchs, CSU, Oct 2014
brfuchs@atmos.colostate.edu

Amendments by
Timothy Lang (tjlangco@gmail.com)
"""

from __future__ import absolute_import
from __future__ import division
import numpy as np
import warnings
from .csu_liquid_ice_mass import (
    calc_zh_zv, calc_zdp, linearize, get_linear_fits)
from .common import (
    _check_for_array, calc_rain_zr, calc_rain_nexrad, calc_rain_kdp_zdr,
    calc_rain_z_zdr, calc_rain_kdp)
from .csu_blended_rain_tropical import calc_blended_rain_tropical


def calc_blended_rain(
        dz=None, zdr=None, kdp=None, ice_flag=False, band='S',
        thresh_dz=38.0, thresh_zdr=0.5, thresh_kdp=0.3,
        thresh_frac_ice=0.1, thresh_nexrad=53.0,
        r_z_a=300.0, r_z_b=1.4, r_kdp_a=40.5, r_kdp_b=0.85,
        r_z_zdr_a=6.7e-3, r_z_zdr_b=0.927, r_z_zdr_c=-0.343,
        fit_a=None, fit_b=None, method='cr1995',
        r_kdp_zdr_a=90.8, r_kdp_zdr_b=0.93, r_kdp_zdr_c=-0.169):

    """
    This algorithm ingests polarimetric radar data and computes rain rate,
    using difference reflectivity as guide for determining presence of ice.

    Inputs:
    dz = Reflectivity
    zdr = Differential Reflectivity
    kdp = Specific Differential Phase
    thresh_dz = Threshold for dz to use certain rain algorithms
    thresh_zdr = Threshold for zdr to use certain rain algorithms
    thresh_kdp = Threshold for kdp to use certain rain algorithms
    thresh_nexrad = Threshold for dz to cap Z-R
    r_z_a, r_z_b => Z = a * R**b
    r_kdp_a, r_kdp_b => R = a * kdp**b
    r_kdp_zdr_a, r_kdp_zdr_b, r_kdp_zdr_c => R = a * kdp**b * 10.0**(c * zdr)
    r_z_zdr_a, r_z_zdr_b, r_z_zdr_c => R = a*linearize(z)**b * 10.0**(c*zdr)
    ice_flag = Set to True to return Zdp and Fraction of Ice
    fit_a, fit_b = Parameters for the rain line
    method = Reference used to set the rain line, if fit_a/b not provided

    Returns: rain rate, method, (Zdp, Fraction of Ice)
    method = 1: R(Kdp, Zdr)
    method = 2: R(Kdp)
    method = 3: R(Z, Zdr)
    method = 4: R(Z)
    method = 5: R(Zrain)

    See Bringi and Chandrasekar textbook for more information.
    """
    # Initialize, check for all vars, check for scalars
    if dz is None or kdp is None or zdr is None:
        warnings.warn('No dz, zdr, or kdp provided, failing ...')
        return
    dz, zdr, kdp, len_flag = _check_for_array(dz, zdr, kdp)
    r_blended = np.zeros_like(dz)
    meth = np.int16(np.zeros_like(dz))

    # Rainfall estimates regardless of frequency
    r_dz_nexrad = calc_rain_nexrad(dz, thresh_nexrad, a=r_z_a, b=r_z_b)
    r_kdp = calc_rain_kdp(kdp, a=r_kdp_a, b=r_kdp_b)

    # Conditions used regardless of frequency
    cond_kdp = kdp >= thresh_kdp
    cond_dz = dz >= thresh_dz
    cond_dz_kdp = np.logical_and(cond_dz, cond_kdp)

    if band == 'S':
        # S-band only rain calculations
        r_kdp_zdr = calc_rain_kdp_zdr(kdp, zdr, a=r_kdp_zdr_a,
                                      b=r_kdp_zdr_b, c=r_kdp_zdr_c)
        r_dz_zdr = calc_rain_z_zdr(dz, zdr, a=r_z_zdr_a,
                                   b=r_z_zdr_b, c=r_z_zdr_c)

        # calculate rain line, zdp, and ice fraction
        zhor, zvert = calc_zh_zv(dz, zdr)
        zdp = calc_zdp(zhor, zvert)

        # calculate contribution to Zh from pure rain
        if fit_a is None:
            fit_a, fit_b = get_linear_fits(method=method)
        zrain = 10.0**((fit_a * zdp + fit_b)/10.0)
        dzrain = 10.0 * np.log10(zrain)
        fi = 1.0 - (zrain / zhor)
        r_dz_rainonly = calc_rain_nexrad(dzrain, thresh_nexrad,
                                         a=r_z_a, b=r_z_b)

        # Additional S-band conditions
        cond_ice = fi < thresh_frac_ice
        cond_zdr = zdr >= thresh_zdr

        # Set of method choices
        cond_meth_1or2 = np.logical_and(cond_ice, cond_dz_kdp)
        cond_meth_1 = np.logical_and(cond_meth_1or2, cond_zdr)
        cond_meth_2a = np.logical_and(cond_meth_1or2, ~cond_zdr)
        cond_meth_2b = np.logical_and(~cond_ice, cond_dz_kdp)
        cond_meth_2 = np.logical_or(cond_meth_2a, cond_meth_2b)
        cond_meth_3or4 = np.logical_and(cond_ice, ~cond_dz_kdp)
        cond_meth_3 = np.logical_and(cond_meth_3or4, cond_zdr)
        cond_meth_4 = np.logical_and(cond_meth_3or4, ~cond_zdr)
        cond_meth_5 = np.logical_and(~cond_ice, ~cond_dz_kdp)

        # Assign methods native only to S-band
        meth[cond_meth_1] = 1
        r_blended[cond_meth_1] = r_kdp_zdr[cond_meth_1]
        meth[cond_meth_3] = 3
        r_blended[cond_meth_3] = r_dz_zdr[cond_meth_3]
        meth[cond_meth_5] = 5
        r_blended[cond_meth_5] = r_dz_rainonly[cond_meth_5]

    else:
        # Logic for other bands; Zdr not used.
        # This will get spoofed by pure hail shaft (high Z, low Kdp)
        # Recommend csu_hidro_rain if hail is possible
        cond_meth_2 = cond_dz_kdp
        cond_meth_4 = ~cond_dz_kdp

    # Assign methods native to all bands
    meth[cond_meth_2] = 2
    r_blended[cond_meth_2] = r_kdp[cond_meth_2]
    meth[cond_meth_4] = 4
    r_blended[cond_meth_4] = r_dz_nexrad[cond_meth_4]

    # Return based on what the user provided and what they wanted
    if not ice_flag or band != 'S':
        if not len_flag:
            return r_blended[0], meth[0]
        else:
            return r_blended, meth
    else:
        if not len_flag:
            return r_blended[0], meth[0], zdp[0], fi[0]
        else:
            return r_blended, meth, zdp, fi


def csu_hidro_rain(
        dz=None, zdr=None, kdp=None, fhc=None, band='S',
        thresh_dz=38.0, thresh_zdr=0.5, thresh_kdp=0.3, thresh_nexrad=53.0,
        r_z_a=300.0, r_z_b=1.4, r_kdp_a=40.5, r_kdp_b=0.85,
        r_z_zdr_a=6.7e-3, r_z_zdr_b=0.927, r_z_zdr_c=-0.343,
        r_kdp_zdr_a=90.8, r_kdp_zdr_b=0.93, r_kdp_zdr_c=-0.169):
    """
    This algorithm ingests polarimetric radar data and computes rain rate,
    using hydrometeor ID as a guide for determining the presence of ice.

    Inputs:
    dz = Reflectivity
    zdr = Differential Reflectivity
    kdp = Specific Differential Phase
    fhc = Hydrometeor ID
    thresh_dz = Threshold for dz to use certain rain algorithms
    thresh_zdr = Threshold for zdr to use certain rain algorithms
    thresh_kdp = Threshold for kdp to use certain rain algorithms
    thresh_nexrad = Threshold for dz to cap Z-R
    r_z_a, r_z_b => Z = a * R**b
    r_kdp_a, r_kdp_b => R = a * kdp**b
    r_kdp_zdr_a, r_kdp_zdr_b, r_kdp_zdr_c => R = a * kdp**b * 10.0**(c * zdr)
    r_z_zdr_a, r_z_zdr_b, r_z_zdr_c => R = a*linearize(z)**b * 10.0**(c*zdr)

    Returns: rain rate, method
    method = 1: R(Kdp, Zdr)
    method = 2: R(Kdp)
    method = 3: R(Z, Zdr)
    method = 4: R(Z)

    See Bringi and Chandrasekar textbook for more information.
    """
    # Initialize, check for all necessary vars, and allow scalars
    if dz is None or kdp is None or zdr is None or fhc is None:
        warnings.warn('No dz, zdr, kdp, or fhc provided, failing ...')
        return
    dz, zdr, kdp, len_flag = _check_for_array(dz, zdr, kdp)
    crr_hidzk = np.zeros_like(dz)
    crr_meth = np.int16(np.zeros_like(dz))

    # Rainfall calculations regardless of frequency
    crr_z88t = calc_rain_nexrad(dz, a=r_z_a, b=r_z_b)
    crr_kd = calc_rain_kdp(kdp, a=r_kdp_a, b=r_kdp_b)

    # Conditions used regardless of frequency
    # fhc = 1, 2, 10 => drizzle, rain, big drops
    cond_rain_a = np.logical_or(fhc == 1, fhc == 2)
    cond_rain = np.logical_or(fhc == 10, cond_rain_a)
    cond_kdp = kdp >= thresh_kdp
    cond_dz = dz >= thresh_dz
    cond_dz_kdp = np.logical_and(cond_dz, cond_kdp)
    # fhc = 5, 7, 8, 9 => wet snow, graupel (LD & HD), hail
    cond_ice_a = np.logical_or(fhc == 5, fhc == 7)
    cond_ice_b = np.logical_or(fhc == 8, fhc == 9)
    cond_ice = np.logical_or(cond_ice_a, cond_ice_b)
    cond_ice_kdp = np.logical_and(cond_ice, cond_kdp)

    if band == 'S':
        # S-band only rain calcs
        crr_zhdr = calc_rain_z_zdr(dz, zdr, a=r_z_zdr_a,
                                   b=r_z_zdr_b, c=r_z_zdr_c)
        crr_kddr = calc_rain_kdp_zdr(kdp, zdr, a=r_kdp_zdr_a,
                                     b=r_kdp_zdr_b, c=r_kdp_zdr_c)

        # S-band only conditions
        cond_zdr = zdr >= thresh_zdr
        cond_dz_kdp_zdr = np.logical_and(cond_dz_kdp, cond_zdr)
        cond_dz_kdp_not_zdr = np.logical_and(cond_dz_kdp, ~cond_zdr)
        cond_not_kdp_zdr = np.logical_and(~cond_kdp, cond_zdr)
        cond_not_kdp_not_zdr = np.logical_and(~cond_kdp, ~cond_zdr)

        # Set methods
        cond_meth_1 = np.logical_and(cond_rain, cond_dz_kdp_zdr)
        cond_meth_2 = np.logical_and(cond_rain, cond_dz_kdp_not_zdr)
        cond_meth_3 = np.logical_and(cond_rain, cond_not_kdp_zdr)
        cond_meth_4a = np.logical_and(cond_rain, cond_not_kdp_not_zdr)
        cond_meth_4b = np.logical_and(cond_rain, ~cond_dz)
        cond_meth_4 = np.logical_or(cond_meth_4a, cond_meth_4b)

        # Assign rain rates based on methods
        crr_meth[cond_meth_1] = 1
        crr_hidzk[cond_meth_1] = crr_kddr[cond_meth_1]
        crr_meth[cond_meth_3] = 3
        crr_hidzk[cond_meth_3] = crr_zhdr[cond_meth_3]
        crr_meth[cond_ice_kdp] = 2
        crr_hidzk[cond_ice_kdp] = crr_kd[cond_ice_kdp]

    else:
        # Logic for other bands; Zdr not used.
        cond_meth_2 = cond_dz_kdp
        cond_meth_4 = np.logical_and(~cond_dz_kdp, cond_rain)
        # (high Z, low Kdp) or (low Z, high Kdp) and no HID rain, R = 0.0
        cond_ice_kdp = np.logical_and(~cond_dz_kdp, ~cond_rain)
        crr_meth[cond_ice_kdp] = 2
        crr_hidzk[cond_ice_kdp] = 0.0

    # Assign methods native to all bands
    crr_meth[cond_meth_2] = 2
    crr_hidzk[cond_meth_2] = crr_kd[cond_meth_2]
    crr_meth[cond_meth_4] = 4
    crr_hidzk[cond_meth_4] = crr_z88t[cond_meth_4]

    if not len_flag:
        return crr_hidzk[0], crr_meth[0]
    else:
        return crr_hidzk, crr_meth



# -*- coding: utf-8 -*-
"""
Python remake of Brenda Dolan's code to calculate water and ice mass.

Brody Fuchs, CSU, Oct 2014
brfuchs@atmos.colostate.edu

Amendments by
Timothy Lang (tjlangco@gmail.com)

Designed around tropical, oceanic equations found in Thompson et al. 2016, and
designed to work with the Powell and Houze 2015 rain typing (or simple
convective /stratiform)
Brenda Dolan (bdolan@atmos.colostate.edu)
"""

from __future__ import absolute_import
from __future__ import division
import numpy as np
import warnings
from .csu_liquid_ice_mass import linearize, get_linear_fits
from .common import (
    _check_for_array, calc_rain_zr, calc_rain_nexrad, calc_rain_kdp_zdr,
    calc_rain_z_zdr, calc_rain_kdp)


def calc_blended_rain_tropical(
        dz=None, zdr=None, kdp=None, cs=None, ice_flag=False,
        fhc=None,
        predef='True', band='S',
        thresh_dz=38.0, thresh_zdr=0.25, thresh_kdp=0.3,
        thresh_frac_ice=0.1, thresh_nexrad=53.0,
        r_z_a=216., r_z_b=1.39,
        r_z_a_c=126., r_z_b_c=1.39, r_z_a_s=291., r_z_b_s=1.55,
        r_kdp_a=59.5202, r_kdp_b=0.7451,
        r_z_zdr_a=0.0085, r_z_zdr_b=0.9237, r_z_zdr_c=-5.2389/10.,
        fit_a=None, fit_b=None, method='cr1995',
        r_kdp_zdr_a=96.5726, r_kdp_zdr_b=0.9315, r_kdp_zdr_c=-2.1140/10.):

    """
    This algorithm ingests polarimetric radar data and computes rain rate,
    based on Thompson et al. 2016. Since ice is not expected at the surface in
    the tropics, the algorithm does not look for ice contamination.

    Inputs:
    dz = Reflectivity
    zdr = Differential Reflectivity
    kdp = Specific Differential Phase
    cs = Convective / Stratiform map (
        3=mixed / uncertain,
        2=convective,
        1= stratiform,
        0=unknown)
    thresh_zdr = Threshold for zdr to use certain rain algorithms
    thresh_kdp = Threshold for kdp to use certain rain algorithms

    Use the predef='True' if you want to use the coefficients defined
    in Thompson 2016. Otherwise specifiy the coefficients for each
    relationship.

    Returns: rain rate, method, (Zdp, Fraction of Ice)
    method = 1: R(Kdp, Zdr)
    method = 2: R(Kdp)
    method = 3: R(Z, Zdr)
    method = 4: R(Z_all)
    method = 5: R(Z_c)
    method = 6: R(Z_s)

    Implemented by B. Dolan July 2016

    See Bringi and Chandrasekar textbook for more information
    See also Thompson et al. 2016 for more details of the Tropical Blended
    algorithm

    Note
    ----

    The coefficients are defined as Z=aR**b .

    The T15 equations are defined as:
    R = a*Kdp*zeta_dr**c and R=a*Zh*zeta_dr**c
    where as the algorithm expects it in the form
    R = a*Kdp*10.**(Zdr*c)  R=a*Zh*10.**(Zdr*c)

    These just differ in the by 1/10. in the exponent of Zdr, so be sure to
    divide by 10. when sending to the algorithm.
    """
    bnd = band
    # Initialize, check for all vars, check for scalars
    if dz is None or kdp is None or zdr is None:
        warnings.warn('No dz, zdr, or kdp provided, failing ...')
        return
    else:
        len_flag = hasattr(dz, '__len__')

        if not len_flag:
            dz = np.array([dz])
            kdp = np.array([kdp])
            zdr = np.array([zdr])

    if predef == 'True':
        """

        """
        # Convective R-Z
        r_z_a_c = 126.
        r_z_b_c = 1.39
        # Stratiform R-Z
        r_z_a_s = 291.
        r_z_b_s = 1.55
        # All R-Z
        r_z_a = 216
        r_z_b = 1.39

        if band == 'S':
            # R = a*Kdp*zeta**c
            r_kdp_zdr_a = 96.5726,
            r_kdp_zdr_b = 0.9315,
            r_kdp_zdr_c = -2.1140/10.,

            # R = a*Zh*zeta**c
            r_z_zdr_a = 0.0085
            r_z_zdr_b = 0.9237
            r_z_zdr_c = -5.2389/10.

            # R = a*Kdp**b
            r_kdp_a = 59.5202
            r_kdp_b = 0.7451

        if band == 'C':
            # R = a*Kdp*zeta**c
            r_kdp_zdr_a = 45.6976
            r_kdp_zdr_b = 0.8763
            r_kdp_zdr_c = -1.6718/10.

            # R = a*Zh*zeta**c
            r_z_zdr_a = 0.0086
            r_z_zdr_b = 0.9088
            r_z_zdr_c = -4.2059/10.

            # R = a*Kdp**b
            r_kdp_a = 34.5703
            r_kdp_b = 0.7331

        if band == 'X':
            # R = a*Kdp*zeta**c
            r_kdp_zdr_a = 28.1289
            r_kdp_zdr_b = 0.9194
            r_kdp_zdr_c = -1.6876/10.

            # R = a*Zh*zeta**c
            r_z_zdr_a = 0.0085
            r_z_zdr_b = 0.9294
            r_z_zdr_c = -4.4580/10.

            # R = a*Kdp**b
            r_kdp_a = 21.9729
            r_kdp_b = 0.7221

    if fhc is None:
        warnings.warn('No FHC ... Rain may be calculated above melting layer')

    r_blended = np.zeros_like(dz)
    # r_blended[...] = -9999.
    meth = np.int16(np.zeros_like(dz)) - 1

    # convective ZR estimate
    r_dz_conv = calc_rain_zr(dz, a=r_z_a_c, b=r_z_b_c)
    # stratiform ZR estimate
    r_dz_strat = calc_rain_zr(dz, a=r_z_a_s, b=r_z_b_s)
    # ALL ZR estimate
    r_dz_all = calc_rain_zr(dz, a=r_z_a, b=r_z_b)

    # Polarimetric estimates
    r_kdp_zdr = calc_rain_kdp_zdr(kdp, zdr, a=r_kdp_zdr_a,
                                  b=r_kdp_zdr_b, c=r_kdp_zdr_c)
    r_dz_zdr = calc_rain_z_zdr(dz, zdr, a=r_z_zdr_a,
                               b=r_z_zdr_b, c=r_z_zdr_c)
    r_kdp = calc_rain_kdp(kdp, a=r_kdp_a, b=r_kdp_b)

    # Conditions
    cond_kdp = kdp >= thresh_kdp
    cond_dbz = dz >= thresh_dz
    cond_kdpz = np.logical_and(cond_kdp,cond_dbz)
    cond_zdr = zdr >= thresh_zdr
    # Set of method choices
    cond_meth_1 = np.logical_and(cond_kdpz, cond_zdr)
    cond_meth_2 = np.logical_and(cond_kdpz, ~cond_zdr)
    cond_meth_3 = np.logical_and(cond_zdr, ~cond_kdpz)
    cond_meth_4 = np.logical_and(~cond_kdpz, ~cond_zdr)

    if cs is None:
        meth[cond_meth_4] = 4
        r_blended[cond_meth_4] = r_dz_all[cond_meth_4]
    else:
        cond_meth_4c = np.logical_and(cond_meth_4, cs == 2)
        cond_meth_4s = np.logical_and(cond_meth_4, cs == 1)
        cond_meth_4a = np.logical_and(cond_meth_4, cs == 3)
        meth[cond_meth_4c] = 5
        meth[cond_meth_4s] = 6
        meth[cond_meth_4a] = 4
        r_blended[cond_meth_4c] = r_dz_conv[cond_meth_4c]
        r_blended[cond_meth_4s] = r_dz_strat[cond_meth_4s]
        r_blended[cond_meth_4a] = r_dz_all[cond_meth_4a]

    # Assign methods
    meth[cond_meth_1] = 1
    meth[cond_meth_2] = 2
    meth[cond_meth_3] = 3
    # Assign rain rates based on methods
    r_blended[cond_meth_1] = r_kdp_zdr[cond_meth_1]
    r_blended[cond_meth_2] = r_kdp[cond_meth_2]
    r_blended[cond_meth_3] = r_dz_zdr[cond_meth_3]

    if fhc is not None:
        cond_ice = np.logical_and(fhc > 2, fhc < 10)
        r_blended[cond_ice] = 0.0
        meth[cond_ice] = -1
        cond_hail = np.logical_and(fhc == 9, cond_kdpz)
        r_blended[cond_hail] = r_kdp[cond_hail]
        meth[cond_ice] = 2

    r_blended[dz < -10] = 0.0
    meth[dz < -10] = -1

    # Return based on what the user provided and what they wanted
    return r_blended, meth



"""
Brenda Dolan CSU, October 2018
bdolan@atmos.colostate.edu

This is a melting layer detection algorithm. It is basically a three category
HID where it is finding the melting layer vs. anything else (other).

There are two different methodologies based on if it is a ppi or rhi scan.

Porting over Elizabeth Thompsons's HID code from IDL.
(Based on Thompson et al. 2014)

Thompson, E. J., Rutledge, S. A., Dolan, B., Chandrasekar, V., & Cheong,
B. L. (2014). A dual-polarization radar hydrometeor classification algorithm
for winter precipitation. Journal of Atmospheric and Oceanic Technology,
31(7), 1457-1481.

(Also based on Dolan and Rutledge, 2009)
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
import numpy as np
import scipy.stats as stats

from .csu_fhc_ml1 import csu_fhc_cold_newml1
from .beta_functions import CSV_DIR


def melting_layer(dz=None, zdr=None, ldr=None, kdp=None, rho=None, sn=None,
                  heights=None, scan_type='ppi', verbose=False,
                  plot_flag=False, band='S', fdir=CSV_DIR, minRH=0.5,
                  expected_ML=None, nsect=36, sn_thresh=5, azimuths=None):
    # The first step ist to run a FHC that determines 'wet snow' from 'other'
    scores = csu_fhc_cold_newml1(
        dz=dz, zdr=zdr, rho=rho, kdp=kdp, use_temp=False, band='S',
        method='linear', verbose=verbose, fdir=fdir)

    fh = np.argmax(scores, axis=0) + 1

    # Now this works a lot better if the SNR is included.
    if sn is not None:
        if verbose:
            print('QCing using SN <', sn_thresh,
                  'to remove edge WS classifications')
        fh[sn < sn_thresh] = 0
    else:
        if verbose:
            print('Wet snow detection will be improved',
                  'by including the SN field.')

    # Now for the ML algorithm Rho is the most important variable.
    # So if the rho was not valid, and where rho is less than the threshold.
    rhfill = rho.filled(fill_value=np.nan)
    whbad = np.where(np.isnan(rhfill))
    fh[whbad] = -1
    if verbose:
        print('Qcing using RH <', minRH)
    fh[rho <= minRH] = -1

    # In this case, fh =1 is 'other', fh = 2 is 'Wet snow'
    # Now use the returned Wet snow field to look at melting layer statistics
    # and try to refine the location of the ML. There are 2 different methods
    # based on if the scan is an rhi or ppi.
    if heights is None:
        print('No heights specified. Returning')
        return
    if scan_type == 'rhi' or scan_type == 'grid':
        if verbose:
            print('scan type is RHI')
        meltlev, mean_melt = get_ml_rhi(
            fh, dz, heights, expected_ML, verbose=verbose)
        print('RADAR MELTING IS:', mean_melt)
    else:
        if verbose:
            print('scan type is PPI')
        meltlev, mean_melt = get_ml_ppi(
            fh, dz, heights, expected_ML, azimuths=azimuths,
            nsect=nsect, verbose=verbose)
        print('RADAR MELTING IS:', np.mean(mean_melt))
    # Lastly, return the HID returned from the WS / other
    # fh[meltlev == 0] = 0
    # fh[meltlev == 2] = 0

    return meltlev, mean_melt, fh, scores


def get_ml_rhi(fh, dz, height, expected_ML, verbose=False):
    # This will return two things. The first is an array the same size as
    # FH data identifying the melting region, the cold and warm regions
    # meltlev = 0   Warm
    # meltlev = 1   Melting Layer
    # meltlev = 2   Cold layer
    # It will also return the height of the melting layer.

    if expected_ML is not None:
        if verbose:
            print('Using an expected ML',
                  'of {m} to narrow the results'.format(m=expected_ML))
        wh_ht = np.ma.logical_and(height <= expected_ML+1.5,
                                  height >= expected_ML-1.5)
        wh_ws = np.ma.logical_and(wh_ht, fh == 2)
        wh_ML = np.ma.where(wh_ws)
    else:
        wh_ML = np.ma.where(fh == 2)
    # find total number of ML pixels within these "proper" limits
    # of height and range
    #print("Checking",np.shape(fh),np.shape(dz))
    
    ML_z = height[wh_ML]
    # define 80th and 20th percentile marks for top and bottom of
    # ML ~ giangrande et al. 2008 / boodoo et al. 2010 methodology
    ML_num = len(ML_z)

    # you are going to have some error messages if it doesn't find a
    # melting layer - it needs an array to do these statistics
    # this algorithm is designed for stratiform precipitation, so if it
    # doesn't have a ML you really should be using a different algorithm
    # because there are no convective precip categories like graupel/hail, etc.
    if len(ML_z) > 10:
        distfreq, eg = np.histogram(ML_z)
        maxfreq = np.max(distfreq)
        ML_mode, ML_count = stats.mode(ML_z)
        ML_sdevZ = np.std(ML_z)
        ML_medianZ = np.median(ML_z)
        ML_varZ = np.var(ML_z)
        ML_skewZ = stats.skew(ML_z)
        ML_kurtZ = stats.kurtosis(ML_z)
        try:
            ML_80Z = np.percentile(ML_z, 80)
        except:
            if verbose:
                print('Not enough points for 80 percent')
            else:
                pass
        try:
            ML_20Z = np.percentile(ML_z, 20)
        except:
            if verbose:
                print('Not enough points for 20 percent')
            else:
                pass
        if verbose:
            print("ML Stats:")
            print("mode: {m:.2f}".format(m=ML_mode[0]))
            print("Median:{o:.2f}".format(o=ML_medianZ))
            print("SDEV: {s:.2f}".format(s=ML_sdevZ))
            print("Var: {v:.2f}".format(v=ML_varZ))
            print("Kurt: {k:.2f}".format(k=ML_kurtZ))
            print("Skew: {w:.2f}".format(w=ML_skewZ))
            print("80per: {e:.2f}".format(e=ML_80Z))
            print("20per: {t:.2f}".format(t=ML_20Z))
        else:
            print("Melting level height:{m:.2f}".format(m=ML_medianZ))
    else:
        print('not enough points for ML detection.',
              'Using expected: {m}'.format(m=expected_ML))
        ML_medianZ = expected_ML

    meltlev = np.zeros_like(fh)+1

    wh_warm = np.where(height < ML_medianZ)
    wh_cold = np.where(height > ML_medianZ)

    meltlev[wh_warm] = 0
    meltlev[wh_cold] = 2
    
    #lastly, return the hid associated with WS and Other
    
    return meltlev,ML_medianZ
    

def get_ml_ppi(fh, dz, height, expected_ML,
               nsect=36, verbose=False, azimuths=None):

    # This is defining variable ML height for each azimuth sector!
    # 36 sectors with 10 deg intervals
    # This will return two things. The first is an array the same size as
    # FH data identifying the melting region, the cold and warm regions
    # meltlev = 0   Warm
    # meltlev = 1   Melting Layer
    # meltlev = 2   Cold layer

    # It will also return the height of the melting layer.
    if azimuths is not None:
        az3d = np.repeat(azimuths[..., np.newaxis],
                         np.shape(height[0, :]), axis=1)
        nsec = nsect
        if verbose:
            print('Number of sections:{n}'.format(n=nsec))
        nint = nsect
        az_sector = np.linspace(0, 360, nint)
        ML_mode = np.zeros([nint-1])
        ML_sdevZ = np.zeros([nint-1])
        ML_medianZ = np.zeros([nint-1])
        ML_varZ = np.zeros([nint-1])
        ML_skewZ = np.zeros([nint-1])
        ML_kurtZ = np.zeros([nint-1])
        ML_80Z = np.zeros([nint-1])
        ML_20Z = np.zeros([nint-1])
        meltlev_bottom_az = np.zeros_like(fh).astype(float)
        meltlev_top_az = np.zeros_like(fh).astype(float)
        meltlev_az = np.zeros_like(fh).astype(float)
        meltlev = np.zeros_like(fh).astype(float) + 1.0

        for i, a in enumerate(az_sector[:-1]):
            wh_az = np.logical_and(az3d >= a, az3d < az_sector[i+1])
            whsec = np.ma.where(wh_az)
            wh_sector1 = np.ma.where(np.logical_and(wh_az, fh == 2))
            wh_ws = np.ma.where(fh[wh_sector1] == 2)
            ML_sector = height[wh_sector1]
            npts = len(np.ravel(whsec))
            # print(npts)
            ML_num_sector = len(ML_sector)
            per_ML = np.float(ML_num_sector)/npts*100.
            # print('ML number in sector', ML_num_sector,
            #       'percent of sector', per_ML)
            # print('Mean dz in ML', np.nanmean(dz[wh_sector1]))
            if (per_ML > 0.05) and (np.nanmean(dz[wh_sector1]) > 15):
                # print('made it to if')
                # print(i, a)
                MLcheck = np.median(ML_sector[wh_ws])
                if verbose:
                    print('ML check, expected ML:',MLcheck, expected_ML)
                if np.abs(MLcheck-expected_ML) > 2.0:
                    print('Melting layer is way off. Double check sounding',
                          'and radar data or consider different HCA.')
                else:
                    ML_mode[i], ML_count = stats.mode(ML_sector[wh_ws])
                    ML_sdevZ[i] = np.std(ML_sector[wh_ws])
                    ML_medianZ[i] = np.median(ML_sector[wh_ws])
                    ML_varZ[i] = np.var(ML_sector[wh_ws])
                    ML_skewZ[i] = stats.skew(ML_sector[wh_ws])
                    ML_kurtZ[i] = stats.kurtosis(ML_sector[wh_ws])
                    ML_80Z[i] = np.percentile(ML_sector[wh_ws], 90)
                    ML_20Z[i] = np.percentile(ML_sector[wh_ws], 10)
            else:
                if verbose:
                    print('Too few points for ML. Consider warm',
                          'HCA or completely winter case.')
            
            whoutside = np.squeeze(
                np.where(np.logical_or(ML_sector <= (ML_20Z[i]),
                                       ML_sector >= (ML_80Z[i]))))
            meltlev_bottom_az[whsec] = ML_20Z[i]
            meltlev_top_az[whsec] = ML_80Z[i]
            meltlev_az[whsec] = ML_medianZ[i]
            if verbose:
                print('Melting layer for {a}-{b}: {m}'.format(
                    a=a, b=az_sector[i+1], m=ML_medianZ[i]))

        wh_warm = np.where(height <= meltlev_az)
        wh_cold = np.where(height >= meltlev_az)
        meltlev[wh_cold] = 2
        meltlev[wh_warm] = 0
        # meltlev[wh_sector1] = meltlevlim
    else:
        print('No azimuths specified. Using RHI version.')
        meltlev, ML_medianZ = get_ml_rhi(fh,dz, height,expected_ML, verbose=False)
    return meltlev, ML_medianZ
